#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import datetime

import numpy as np
import pandas as pd

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def datetime2iso8601(time):
    r"""Transforms datetime to TT2000 string format.

    Parameters
    ----------
    time : datetime.datetime
        Time to convert to tt2000 string.

    Returns
    -------
    tt2000 : str
        Time in TT20000 iso_8601 format.

    """

    # Check input type
    message = "time must be array_like or datetime.datetime"
    assert isinstance(time, (list, np.ndarray, datetime.datetime)), message

    if isinstance(time, (np.ndarray, list)):
        return list(map(datetime2iso8601, time))

    assert isinstance(time, datetime.datetime), "time datetime.datetime"

    time_datetime = pd.Timestamp(time)

    # Convert to string
    datetime_str = time_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")

    time_iso8601 = f"{datetime_str}{time_datetime.nanosecond:03d}"

    return time_iso8601



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import string

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def make_labels(axs, pos, pad: float = 0, **kwargs):
    r"""Add subplots labels to axes

    Parameters
    ----------
    axs : ndarray
        Array of subplots axes.
    pos : array_like
        Position of the text in the axis.

    Returns
    -------
    axs : ndarray
        Array of subplots axes with labels.

    """

    lbl = string.ascii_lowercase[pad : len(axs) + pad]

    for label, axis in zip(lbl, axs):
        if "proj" in axis.properties():
            axis.text2D(
                pos[0],
                pos[1],
                f"({label})",
                transform=axis.transAxes,
                **kwargs,
            )
        else:
            axis.text(
                pos[0],
                pos[1],
                f"({label})",
                transform=axis.transAxes,
                **kwargs,
            )

    return axs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-on imports
import os
import warnings

# 3rd party imports
import numpy as np
import pandas as pd
from scipy import interpolate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def igrf(time, flag):
    r"""Returns magnetic dipole latitude and longitude of the IGRF model

    Parameters
    ----------
    time : numpy.ndarray
        Times in unix format.
    flag : str
        Default is dipole.


    Returns
    -------
    lambda : ndarray
        latitude

    phi : ndarray
        longitude

    """

    # Root path
    # root_path = os.getcwd()
    path = os.sep.join(
        [os.path.dirname(os.path.abspath(__file__)), "igrf13coeffs.csv"],
    )
    df = pd.read_csv(path)

    # construct IGRF coefficient matrices
    years_igrf = df.loc[0][3:].to_list()
    years_igrf[-1] = float(years_igrf[-1].split("-")[0]) + 5.0

    # read in all IGRF coefficients from file
    i_igrf = df.loc[1:,].values
    i_igrf[:, -1] = i_igrf[:, -2] + 5.0 * i_igrf[:, -1].astype(np.float64)
    h_igrf = i_igrf[i_igrf[:, 0] == "h", 1:].astype(np.float64)
    g_igrf = i_igrf[i_igrf[:, 0] == "g", 1:].astype(np.float64)

    # timeVec = irf_time(t,'vector');
    # yearRef = timeVec(:,1);
    year_ref = (time * 1e9).astype("datetime64[ns]")
    year_ref = year_ref.astype("datetime64[Y]")
    year_ref = year_ref.astype(np.int64) + 1970
    year_ref_unix = (year_ref - 1970).astype("datetime64[Y]")
    year_ref_unix = year_ref_unix.astype("datetime64[ns]").astype(np.int64) / 1e9

    if np.min(year_ref) < np.min(years_igrf):
        message = (
            "requested time is earlier than the first available IGRF "
            "model from extrapolating in past.. "
        )
        warnings.warn(message, category=UserWarning)

    assert flag == "dipole", "input flag is not recognized"

    tck_g0_igrf = interpolate.interp1d(
        years_igrf,
        g_igrf[0, 2:],
        kind="linear",
        fill_value="extrapolate",
    )
    tck_g1_igrf = interpolate.interp1d(
        years_igrf,
        g_igrf[1, 2:],
        kind="linear",
        fill_value="extrapolate",
    )
    tck_h0_igrf = interpolate.interp1d(
        years_igrf,
        h_igrf[0, 2:],
        kind="linear",
        fill_value="extrapolate",
    )

    g01 = tck_g0_igrf(year_ref + (time - year_ref_unix) / (365.25 * 86400))
    g11 = tck_g1_igrf(year_ref + (time - year_ref_unix) / (365.25 * 86400))
    h11 = tck_h0_igrf(year_ref + (time - year_ref_unix) / (365.25 * 86400))
    lambda_ = np.arctan(h11 / g11)
    phi = np.pi / 2
    phi -= np.arcsin((g11 * np.cos(lambda_) + h11 * np.sin(lambda_)) / g01)

    return np.rad2deg(lambda_), np.rad2deg(phi)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants, integrate

# Local imports
from ..pyrf.resample import resample
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eis_moments(
    inp,
    specie: str = "proton",
    n_bg: xr.DataArray = None,
    p_bg: xr.DataArray = None,
):
    r"""Computes the partial moments given the omni-directional differential
    particle flux and the ion specie under the assumption of angular isotropy
    and non-relativistic ions using the formula from [1]_

    .. math::

        n \\left [ m^{-3} \\right ] = 4 \\pi \\sqrt{\\frac{m_i}{2}} \\sum_{i}
        \\left ( E_i^{1/2} \\right)^0 \\left ( \\frac{J_i}{E_i}\\right)
        \\left ( E_i^{1/2} \\textrm{d} E_i\\right)

        P \\left [ Pa \\right ] = 4 \\pi \\sqrt{\\frac{m_i}{2}} \\sum_{i}
        \\left ( E_i^{1/2} \\right)^2 \\left ( \\frac{J_i}{E_i}\\right)
        \\left ( E_i^{1/2} \\textrm{d} E_i\\right)

        T \\left [ K \\right ] = \\frac{P}{n k_b}

    Parameters
    ----------
    inp : xarray.DataArray
        Omni-directional differential particle flux.
    specie : {"proton", "alpha", "oxygen"}, Optional
        Particle specie. Default is "proton".
    n_bg : xarray.DataArray, Optional
        Time series of the background density. If None do not remove
        penetrating radiations.
    p_bg : xarray.DataArray, Optional
        Time series of the background pressure. If None do not remove the
        penetrating radiations.

    Returns
    -------
    n : xarray.DataArray
        Time series of the number density in [cm^{-3}]
    p : xarray.DataArray
        Time series of the pressure in [nPa]
    t : xarray.DataArray
        Time series of the temperature in [eV]

    Notes
    -----
    The input omni-directional differential particle flux must be given in
    [(1/cm^2 s sr keV)^{-1}], and the energy must be in [keV].
    The integration is performed using the composite Simpson’s rule.

    References
    ----------
    .. [1]  Mauk, B. H., D. G. Mitchell, R. W. McEntire, C. P. Paranicas,
            E. C. Roelof, D. J. Williams, S. M. Krimigis, and A. Lagg (2004),
            Energetic ion characteristics and neutral gas interactions in
            Jupiter’s magnetosphere, J. Geophys. Res., 109, A09S12,
            doi:10.1029/2003JA010270.

    """

    assert specie in ["proton", "alpha", "oxygen"]

    if specie == "proton":
        mass = constants.proton_mass
    elif specie == "alpha":
        mass = constants.proton_mass * 4
    else:
        mass = constants.proton_mass * 16

    # Convert energy and differential particle flux to SI units keV -> J and
    # 1/(cm^2 s sr keV) -> 1/(m^2 s sr J)
    energy = 1e3 * constants.electron_volt * inp.energy.data.copy()
    intensity = 1e4 * inp.data.copy() / (1e3 * constants.electron_volt)

    fact = 4 * np.pi * np.sqrt(mass / 2)

    # Define the integrand as a function of the differential particle flux,
    # the energy and the moment order
    def _int(flux_c, energy_c, order):
        return np.sqrt(energy_c) ** order * (flux_c / energy_c) * np.sqrt(energy_c)

    # Zeroth order moment number density m^-3
    n_i = fact * integrate.simps(_int(intensity, energy, 0), energy, axis=1)

    # First order moment bulk velocity m s^-1
    v_i = fact * integrate.simps(_int(intensity, energy, 1), energy, axis=1)
    v_i /= n_i * np.sqrt(mass / 2)
    # Second order moment pressure in N m^-2 (Pa)
    p_i = fact * integrate.simps(_int(intensity, energy, 2), energy, axis=1)
    t_i = p_i / (n_i * constants.Boltzmann)  # K

    # Convert to usual units
    n_i = ts_scalar(inp.time.data, n_i * 1e-6)  # cm^-3
    v_i = ts_scalar(inp.time.data, v_i * 1e-3)  # km s^-1
    p_i = ts_scalar(inp.time.data, p_i * 1e9)  # nPa
    t_i = ts_scalar(
        inp.time.data,
        t_i * constants.Boltzmann / constants.elementary_charge,
    )  # eV

    if n_bg is not None and p_bg is not None:
        # Resample background to differential particle flux sampling
        n_bg = resample(n_bg, n_i)
        p_bg = resample(p_bg, p_i)

        # Remove background density
        n_i.data -= n_bg.data

        # Remove background pressure
        p_i.data -= 3.0 * p_bg
        p_i.data += constants.proton_mass * n_i * v_i**2 * (n_bg / (n_i - n_bg))

        # Update temperature
        t_i.data = 1e-9 * p_i.data / (1e6 * n_i.data * constants.electron_volt)

    return n_i, v_i, p_i, t_i



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .ts_skymap import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def dist_append(inp0, inp1):
    r"""Concatenate two distribution skymaps along the time axis.

    Parameters
    ----------
    inp0 : xarray.Dataset
        3D skymap distribution at early times.
    inp1 : xarray.Dataset
        3D skymap distribution at late times.

    Returns
    -------
    out : xarray.Dataset
        3D skymap of the concatenated 3D skymaps.

    Notes
    -----
    The time series have to be in the correct time order.

    """

    # Check input type
    assert isinstance(inp1, xr.Dataset), "inp1 must be a xarray.Dataset"

    if inp0 is None:
        return inp1

    assert isinstance(inp0, xr.Dataset), "inp0 must be a xarray.Dataset"

    # Global attributes
    glob_attrs = inp0.attrs

    # Stack coordinates
    # time
    time = np.hstack([inp0.time.data, inp1.time.data])

    # Azimuthal angle
    # if inp0.phi.ndim == 2:
    #    phi = np.vstack([inp0.phi.data, inp1.phi.data])
    # else:
    #    phi = inp0.phi.data
    phi = np.vstack([inp0.phi.data, inp1.phi.data])

    # Elevation angle
    theta = inp0.theta.data

    # Coordinates attributes
    coords_attrs = {k: inp0[k].attrs for k in ["time", "energy", "phi", "theta"]}

    # distribution
    data = np.vstack([inp0.data, inp1.data])
    data_attrs = inp0.data.attrs

    if "delta_energy_plus" in glob_attrs:
        delta_energy_plus = np.vstack(
            [
                inp0.attrs["delta_energy_plus"].data,
                inp1.attrs["delta_energy_plus"].data,
            ],
        )
        glob_attrs["delta_energy_plus"] = delta_energy_plus

    if "delta_energy_minus" in glob_attrs:
        delta_energy_minus = np.vstack(
            [
                inp0.attrs["delta_energy_minus"].data,
                inp1.attrs["delta_energy_minus"].data,
            ],
        )
        glob_attrs["delta_energy_minus"] = delta_energy_minus

    step_table = np.hstack(
        [inp0.attrs["esteptable"], inp1.attrs["esteptable"]],
    )

    energy = np.vstack([inp0.energy.data, inp1.energy.data])

    out = ts_skymap(
        time,
        data,
        energy,
        phi,
        theta,
        energy0=inp0.energy0,
        energy1=inp0.energy1,
        esteptable=step_table,
        attrs=data_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants, optimize

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _disprel(w, *args):
    assert len(args) == 6, "not enougth arguments"
    k, theta, v_a, c_s, wc_e, wc_p = args

    theta = np.deg2rad(theta)
    l_00 = 1
    l_01 = -(w**2) / (k**2 * v_a**2)
    l_02 = -(w**2) / (wc_e * wc_p)
    l_03 = k**2 * np.sin(theta) ** 2 / ((w / c_s) ** 2 - k**2)
    l_0_ = l_00 + l_01 + l_02 + l_03

    l_10 = np.cos(theta) ** 2
    l_11 = -(w**2) / (k**2 * v_a**2)
    l_12 = -(w**2) / (wc_e * wc_p)
    l_1_ = l_10 + l_11 + l_12

    r_0_ = w**2 * np.cos(theta) ** 2 / wc_p**2

    disprel = l_0_ * l_1_ - r_0_

    return disprel


def one_fluid_dispersion(b_0, theta, ions, electrons, n_k: int = 100):
    r"""Solves the one fluid dispersion relation.

    Parameters
    ----------
    b_0 : float
        Magnetic field

    theta : float
        The angle of propagation of the wave with respect to the magnetic
        field, :math:`\cos^{-1}(k_z / k)`

    ions : dict
        Hash table with n : number density, t: temperature, gamma:
        polytropic index.

    electrons : dict
        Hash table with n : number density, t: temperature, gamma:
        polytropic index.

    n_k : int, optional
        Number of wavenumbers.

    Returns
    -------
    wc_1 : xarray.DataArray
        1st root

    wc_2 : xarray.DataArray
        2nd root

    wc_3 : xarray.DataArray
        3rd root

    """

    keys = ["n", "t", "gamma"]
    n_p, t_p, gamma_p = [ions[k] for k in keys]
    n_e, t_e, gamma_e = [electrons[k] for k in keys]

    q_e = constants.elementary_charge
    m_e = constants.electron_mass
    m_p = constants.proton_mass
    ep_0 = constants.epsilon_0
    mu_0 = constants.mu_0

    wc_e = q_e * b_0 / m_e
    wc_p = q_e * b_0 / m_p

    wp_e = np.sqrt(q_e**2 * n_e / (ep_0 * m_e))
    wp_p = np.sqrt(q_e**2 * n_p / (ep_0 * m_p))

    v_p = np.sqrt(q_e * t_p / m_p)
    v_e = np.sqrt(q_e * t_e / m_e)

    v_a = b_0 / np.sqrt(mu_0 * n_p * m_p)
    c_s = np.sqrt((gamma_e * q_e * t_e + gamma_p * q_e * t_p) / (m_e + m_p))

    k_vec = np.linspace(2e-7, 1.0e-4, n_k)

    wc_1, wc_2, wc_3 = [np.zeros(len(k_vec)) for _ in range(3)]

    for i, k in enumerate(k_vec):
        if i < 10:
            guess_w1 = v_a * k * 1.50
            guess_w2 = v_a * k * 0.70
            guess_w3 = c_s * k * 0.99
        else:
            guess_w1 = wc_1[i - 1] + (wc_1[i - 1] - wc_1[i - 2])
            guess_w2 = wc_2[i - 1] + (wc_2[i - 1] - wc_2[i - 2])
            guess_w3 = wc_3[i - 1] + (wc_3[i - 1] - wc_3[i - 2])

        arguments = (k, theta, v_a, c_s, wc_e, wc_p)
        wc_1[i] = optimize.fsolve(_disprel, guess_w1, args=arguments)[0]
        wc_2[i] = optimize.fsolve(_disprel, guess_w2, args=arguments)[0]
        wc_3[i] = optimize.fsolve(_disprel, guess_w3, args=arguments)[0]

    attrs = {
        "wc_e": wc_e,
        "wc_p": wc_p,
        "wp_e": wp_e,
        "wp_p": wp_p,
        "v_p": v_p,
        "v_e": v_e,
        "v_a": v_a,
        "c_s": c_s,
    }

    wc_1 = xr.DataArray(wc_1, coords=[k_vec], dims=["k"], attrs=attrs)
    wc_2 = xr.DataArray(wc_2, coords=[k_vec], dims=["k"], attrs=attrs)
    wc_3 = xr.DataArray(wc_3, coords=[k_vec], dims=["k"], attrs=attrs)

    return wc_1, wc_2, wc_3



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .resample import resample
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eb_nrf(e_xyz, b_xyz, v_xyz, flag=0):
    """Find E and B in MP system given B and MP normal vector.

    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.
    b_xyz : xarray.DataArray
        Time series of the magnetic field.
    v_xyz : xarray.DataArray
        Normal vector.
    flag : str or ndarray
        Method flag :
        * a : L is along b_xyz, N closest to v_xyz and M = NxL
        * b : N is along v_xyz, L is the mean direction of b_xyz in plane perpendicular
        to N, and M = NxL
        * numpy,ndarray : N is along v_xyz , L is closest to the direction specified by
        L_vector (e.g., maximum variance direction), M = NxL

    Returns
    -------
    out : xarray.DataArray
        to fill.

    """
    # Check inputs
    assert isinstance(e_xyz, xr.DataArray), "e_xyz must be a xarray.DataArray"
    assert isinstance(b_xyz, xr.DataArray), "b_xyz must be a xarray.DataArray"
    assert isinstance(v_xyz, xr.DataArray), "v_xyz must be a xarray.DataArray"

    assert e_xyz.ndim == 2 and e_xyz.shape[1] == 3, "e_xyz must be a vector"
    assert b_xyz.ndim == 2 and b_xyz.shape[1] == 3, "e_xyz must be a vector"
    assert v_xyz.ndim == 2 and v_xyz.shape[1] == 3, "e_xyz must be a vector"

    assert isinstance(flag, (str, np.ndarray, list)), "Invalid flag type"

    if isinstance(flag, str):
        assert flag.lower() in ["a", "b"], "flag must be a or b"
        flag_case = flag
        l_direction = None

    else:
        flag = np.array(flag)
        assert flag.ndim == 1 and len(flag) == 3, "array_like flag must be a vector!"
        l_direction = flag
        flag_case = "c"

    if flag_case == "a":
        b_xyz = resample(b_xyz, e_xyz)

        n_l = b_xyz.data / np.linalg.norm(b_xyz.data, axis=1, keepdims=True)
        n_n = np.cross(np.cross(b_xyz.data, v_xyz.data), b_xyz.data)
        n_n /= np.linalg.norm(n_n, axis=1, keepdims=True)
        n_m = np.cross(n_n, n_l)  # in (vn x b) direction
    elif flag_case == "b":
        n_n = v_xyz.data / np.linalg.norm(v_xyz, axis=1, keepdims=True)
        n_m = np.cross(n_n, np.mean(b_xyz.data, axis=0))
        n_m /= np.linalg.norm(n_m, axis=1, keepdims=True)
        n_l = np.cross(n_m, n_n)

    else:
        n_n = v_xyz.data / np.linalg.norm(v_xyz, axis=1, keepdims=True)
        n_m = np.cross(n_n, l_direction)
        n_m /= np.linalg.norm(n_m, axis=1, keepdims=True)
        n_l = np.cross(n_m, n_n)

    # estimate e in new coordinates
    e_lmn = np.vstack([np.sum(e_xyz.data * vec, axis=1) for vec in [n_l, n_m, n_n]])
    out = ts_vec_xyz(e_xyz.time.data, np.transpose(e_lmn), e_xyz.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def struct_func(inp, scales, order):
    r"""Returns the structure function of a time series

    .. math::

       y= \frac{1}{N-s}\sum_{i=1}^{N-s}(x_i - x_{i+s})^o

    where :math:`s` is the scale, and :math:`o` is the order.

    Parameters
    ----------
    inp : xarray.DataArray
        Input time series.
    scales : array_like
        A list or an array containing the scales to calculate.
    order : int
        Order of the exponential of the structure function.
    ncut : int, Optional
        Number of standard deviation to cut (Kiyani et al., XXXX)
    Returns
    -------
    values : xarray.DataArray
        An xarray containing the structure functions, one per product in
        the original time series. The index coordinate contains the scale
        value, and the attribute 'order' keeps a record on the order used
        for its calculation.

    """

    if scales is None:
        scales = np.arange(1, len(inp) // 2)

    if inp.ndim == 1:
        data = inp.data[:, np.newaxis]
    else:
        data = inp.data

    result = []
    for scale in scales:
        increment = np.abs(data[scale:, ...] - data[:-scale, ...])
        result.append(np.nanmean(increment**order, axis=0))

    _, *comp = [inp.coords[dim].data for dim in inp.dims]

    result = xr.DataArray(
        np.squeeze(np.stack(result)),
        coords=[scales, *comp],
        dims=["scales", *inp.dims[1:]],
        attrs=inp.attrs,
    )

    result.attrs["order"] = order

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv

# Built-in imports
import os

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf.datetime642unix import datetime642unix
from ..pyrf.iso86012datetime64 import iso86012datetime64
from ..pyrf.unix2datetime64 import unix2datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def read_feeps_sector_masks_csv(tint):
    r"""Reads the FEEPS sectors to mask due to sunlight contamination from
    csv files.x

    Parameters
    ----------
    tint : list of str
        time range of interest [starttime, endtime] with the format
        "YYYY-MM-DD", "YYYY-MM-DD" or to specify more or less than a day [
        'YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

    Returns
    -------
    mask : dict
        Hash table containing the sectors to mask for each spacecraft and
        sensor ID

    """

    masks = {}

    dates = [
        1447200000.0000000,  # 11/11/2015
        1468022400.0000000,  # 7/9/2016
        1477612800.0000000,  # 10/28/2016
        1496188800.0000000,  # 5/31/2017
        1506988800.0000000,  # 10/3/2017
        1538697600.0000000,
    ]  # 10/5/2018

    # find the file closest to the start time
    date = datetime642unix(iso86012datetime64(np.array(tint)[0]))
    nearest_date = dates[np.argmin((np.abs(np.array(dates) - date)))]
    nearest_date = unix2datetime64(np.array(nearest_date))
    str_date = str(nearest_date.astype("<M8[D]"))
    str_date = str_date.replace("-", "")

    for mms_sc in np.arange(1, 5):
        file_name = f"MMS{mms_sc:d}_FEEPS_ContaminatedSectors_{str_date}.csv"
        csv_file = os.sep.join(
            [os.path.dirname(os.path.abspath(__file__)), "sun", file_name],
        )

        with open(csv_file, "r", encoding="utf-8") as csv_file:
            csv_data = [[float(x) for x in line] for line in csv.reader(csv_file)]

        csv_data = np.array(csv_data)

        for i in range(0, 12):
            mask_vals = []
            for val_idx in range(len(csv_data[:, i])):
                if csv_data[val_idx, i] == 1:
                    mask_vals.append(val_idx)

            masks[f"mms{mms_sc:d}_imask_top-{i + 1:d}"] = mask_vals

        for i in range(0, 12):
            mask_vals = []

            for val_idx in range(len(csv_data[:, i + 12])):
                if csv_data[val_idx, i + 12] == 1:
                    mask_vals.append(val_idx)

            masks[f"mms{mms_sc:d}_imask_bottom-{i + 1:d}"] = mask_vals

    return masks



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _idx_closest(lst0, lst1):
    return [(np.abs(np.asarray(lst0) - k)).argmin() for k in lst1]


def eis_spec_combine_sc(omni_vars, method: str = "mean"):
    r"""Combines omni-directional energy spectrogram variable from EIS on
    multiple MMS spacecraft.

    Parameters
    ----------
    omni_vars : list of xarray.DataArray
        Omni-directional energy spectrograms of all spacecraft.
    method : str, Optional
        Method to combine spectra, "mean" or "sum"

    Returns
    -------
    omni_spec : xarray.DataArray
        Combined omni-directional energy spectrogram.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_omni, pyrfu.mms.eis_pad_combine_sc

    Examples
    --------
    >>> from pyrfu.mms import get_eis_allt, eis_omni, eis_spec_combine_sc

    Define time interval

    >>> tint = ["2017-07-23T16:10:00", "2017-07-23T18:10:00"]

    Load EIS ExTOF flux spectrograms for all 6 telescopes for all spacecraft

    >>> extof_allt_mms = []
    >>> for ic in range(2, 5):
    ...     extof_allt_mms.append(get_eis_allt("flux_extof_proton_srvy_l2",
    ...                             tint, ic))

    Compute the omni-direction flux spectrogram for all spacecraft

    >>> extof_omni_mms = []
    >>> for extof_allt in extof_allt_mms:
    ...     extof_omni_mms.append(eis_omni(extof_allt))

    Combine spectrograms of all spacecraft

    >>> extof_omni_mmsx = eis_spec_combine_sc(extof_omni_mms)

    """

    assert method.lower() in ["mean", "sum"]

    reftime_sc_loc = np.argmin([len(x_.time.data) for x_ in omni_vars])
    refener_sc_loc = np.argmin([len(x_.energy.data) for x_ in omni_vars])

    time_refprobe = omni_vars[reftime_sc_loc]
    ener_refprobe = omni_vars[refener_sc_loc]

    nt_ref, ne_ref = [time_refprobe.shape[0], ener_refprobe.shape[1]]

    # time x energy x spacecraft
    omni_spec_data = np.empty([nt_ref, ne_ref, len(omni_vars)])
    omni_spec_data[:] = np.nan
    # time x energy
    omni_spec = np.empty([nt_ref, ne_ref])
    omni_spec[:] = np.nan

    # energy_data = np.zeros([ne_ref, len(omni_vars)])

    # Average omni flux over all spacecraft and define common energy grid
    for i_p, flux_ in enumerate(omni_vars):
        idx_closest = _idx_closest(
            flux_.energy.data,
            ener_refprobe.energy.data,
        )
        # energy_data[:, pp] = flux_.energy.data[0:ne_ref]
        omni_spec_data[0:nt_ref, :, i_p] = flux_.data[0:nt_ref, idx_closest]

    # Average omni flux over all spacecraft
    for i_t, i_e in itertools.product(range(nt_ref), range(ne_ref)):
        if method.lower() == "mean":
            omni_spec[i_t, i_e] = np.nanmean(
                omni_spec_data[i_t, i_e, :],
                axis=0,
            )
        else:
            omni_spec[i_t, i_e] = np.nansum(
                omni_spec_data[i_t, i_e, :],
                axis=0,
            )

    omni_spec = xr.DataArray(
        omni_spec,
        coords=[time_refprobe.time.data, ener_refprobe.energy.data],
        dims=["time", "energy"],
    )
    omni_spec.attrs["energy_dplus"] = ener_refprobe.attrs["energy_dplus"]
    omni_spec.attrs["energy_dminus"] = ener_refprobe.attrs["energy_dminus"]

    return omni_spec



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def time_clip(inp, tint):
    r"""Time clip the input (if time interval is TSeries clip between start
    and stop).

    Parameters
    ----------
    inp : xarray.DataArray or xarray.Dataset
        Time series of the quantity to clip.
    tint : xarray.DataArray or ndarray or list
        Time interval can be a time series, a array of datetime64 or a list.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time clipped input.

    """

    if isinstance(inp, xr.Dataset):
        coords_data = [inp[k] for k in filter(lambda x: x != "time", inp.dims)]
        coords_data = [time_clip(inp.time, tint), *coords_data]
        out_dict = {dim: coords_data[i] for i, dim in enumerate(inp.coords)}

        for k in inp:
            if "time" in list(inp[k].coords):
                out_dict[k] = time_clip(inp[k], tint)
            else:
                out_dict[k] = inp[k]

        # Find array_like attributes
        arr_attrs = filter(
            lambda x: isinstance(inp.attrs[x], np.ndarray),
            inp.attrs,
        )
        arr_attrs = list(arr_attrs)

        # Initialize attributes dictionary with non array_like attributes
        gen_attrs = filter(lambda x: x not in arr_attrs, inp.attrs)
        out_attrs = {k: inp.attrs[k] for k in list(gen_attrs)}

        for a in arr_attrs:
            attr = inp.attrs[a]

            # If array_like attributes have one dimension equal to time
            # length assume time dependent. One option would be move the time
            # dependent array_like attributes to time series to zVaraibles to
            # avoid confusion
            if attr.shape[0] == len(inp.time.data):
                coords = [np.arange(attr.shape[i + 1]) for i in range(attr.ndim - 1)]
                dims = [f"idx{i:d}" for i in range(attr.ndim - 1)]
                attr_ts = xr.DataArray(
                    attr,
                    coords=[inp.time.data, *coords],
                    dims=["time", *dims],
                )
                out_attrs[a] = time_clip(attr_ts, tint).data
            else:
                out_attrs[a] = attr

        out_attrs = {k: out_attrs[k] for k in sorted(out_attrs)}

        out = xr.Dataset(out_dict, attrs=out_attrs)

        return out

    if isinstance(tint, xr.DataArray):
        t_start, t_stop = tint.time.data[[0, -1]]
    elif isinstance(tint, (np.ndarray, list)):
        if isinstance(tint[0], np.datetime64):
            t_start, t_stop = tint
        elif isinstance(tint[0], str):
            t_start, t_stop = iso86012datetime64(np.array(tint))
        else:
            raise TypeError("Values must be in datetime64, or str!!")
    else:
        raise TypeError("tint must be a DataArray or array_like!!")

    idx_min = bisect.bisect_left(inp.time.data, t_start)
    idx_max = bisect.bisect_right(inp.time.data, t_stop)

    coords = [inp.time.data[idx_min:idx_max]]
    coords_attrs = [inp.time.attrs]

    if len(inp.coords) > 1:
        for k in inp.dims[1:]:
            coords.append(inp.coords[k].data)
            coords_attrs.append(inp.coords[k].attrs)

    out = xr.DataArray(
        inp.data[idx_min:idx_max, ...],
        coords=coords,
        dims=inp.dims,
        attrs=inp.attrs,
    )

    for i, k in enumerate(inp.dims):
        out[k].attrs = coords_attrs[i]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def calc_dng(p_xyz):
    r"""Computes agyrotropy coefficient as in [15]_

    .. math::

        D_{ng} = \frac{\sqrt{8 (P_{12}^2 + P_{13}^2 + P_{23}^2)}}
        {P_\parallel + 2 P_\perp}


    Parameters
    ----------
    p_xyz : xarray.DataArray
        Time series of the pressure tensor

    Returns
    -------
    d_ng : xarray.DataArray
        Time series of the agyrotropy coefficient of the specie.

    References
    ----------
    .. [15] Aunai, N., M. Hesse, and M. Kuznetsova (2013), Electron
            nongyrotropy in the context of collisionless magnetic
            reconnection, Phys. Plasmas, 20(6), 092903,
            doi: https://doi.org/10.1063/1.4820953.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> ic = 1

    Load magnetic field and electron pressure tensor

    >>> b_xyz = mms.get_data("b_gse_fgm_srvy_l2", tint, 1)
    >>> p_xyz_e = mms.get_data("pe_gse_fpi_fast_l2", tint, 1)

    Rotate electron pressure tensor to field aligned coordinates

    >>> p_fac_e_pp = mms.rotate_tensor(p_xyz_e, "fac", b_xyz, "pp")

    Compute agyrotropy coefficient

    >>> d_ng_e = pyrf.calc_dng(p_fac_e_pp)

    """

    # Check input type
    assert isinstance(p_xyz, xr.DataArray), "p_xyz must be a xarray.DataArray"

    # Check import shape
    message = "p_xyz must be a time series of a tensor"
    assert p_xyz.data.ndim == 3 and p_xyz.shape[1] == 3 and p_xyz.shape[2] == 3, message

    # Parallel and perpendicular components
    p_para = p_xyz.data[:, 0, 0]
    p_perp = (p_xyz.data[:, 1, 1] + p_xyz.data[:, 2, 2]) / 2

    # Off-diagonal terms
    p_12, p_13, p_23 = [p_xyz.data[:, 0, 1], p_xyz.data[:, 0, 2], p_xyz.data[:, 1, 2]]

    d_ng = np.sqrt(8 * (p_12**2 + p_13**2 + p_23**2))
    d_ng /= p_para + 2 * p_perp
    d_ng = ts_scalar(p_xyz.time.data, d_ng)

    return d_ng



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _estimate_capa_disk(radius):
    return 8 * constants.epsilon_0 * radius


def _estimate_capa_sphe(radius):
    return 4 * np.pi * constants.epsilon_0 * radius


def _estimate_capa_wire(radius, length):
    if length and radius != 0 and length >= 10 * radius:
        l_ = np.log(length / radius)
        out = length / l_ * (1 + 1 / l_ * (1 - np.log(2)))
        out *= 2 * np.pi * constants.epsilon_0
    else:
        raise ValueError(
            "capacitance_wire requires length at least 10 times the radius!",
        )
    return out


def _estimate_capa_cyli(a, h):
    coef = 4 * np.pi**2 * a * constants.epsilon_0

    if 0.5 < h / a < 4:
        c_1 = h / (2.0 * a * (np.log(16 * h / a) ** 2 + np.pi**2 / 12))
        out = coef * np.pi * c_1
    elif h / a >= 4:
        o_m = 2 * (np.log(4 * h / a) - 1)
        c_1 = 2 * h / (np.pi * a) * (1.0 / o_m + (4 - np.pi**2) / o_m**3)
        out = coef * c_1
    else:
        raise ValueError("length less than diameter, do not have formula yet")

    return out


def estimate(what_to_estimate: str, radius: float, length: float = None):
    r"""Estimate values for some everyday stuff.

    Parameters
    ----------
    what_to_estimate : str
        Value to estimate:
            * "capacitance_disk" estimates the capacitance of a disk
            (requires radius of the disk).
            * "capacitance_sphere" estimates of a sphere
            (requires radius of the sphere).
            * "capacitance_wire" estimates the capacitance of a wire
            (requires radius and length of the wire).
            * "capacitance_cylinder" estimates the capacitance of a cylinder
            (requires radius and half length of the cylinder).
    radius :  float
        Radius of the disk, sphere, wire or cylinder
    length : float, Optional
        Length of the wire or half lenght of the cylinder.

    Returns
    -------
    out : float
        Estimated value.

    Examples
    --------
    >>> from pyrfu import pyrf

    Define radius of the sphere in SI units

    >>> r_sphere = 20e-2

    Computes the capacitance of the sphere

    >>> c_sphere = pyrf.estimate("capacitance_sphere", r_sphere)


    """
    if what_to_estimate.lower() == "capacitance_disk":
        out = _estimate_capa_disk(radius)
    elif what_to_estimate.lower() == "capacitance_sphere":
        out = _estimate_capa_sphe(radius)
    elif what_to_estimate.lower() == "capacitance_wire":
        out = _estimate_capa_wire(radius, length)
    elif what_to_estimate.lower() == "capacitance_cylinder":
        out = _estimate_capa_cyli(radius, length)
    else:
        raise NotImplementedError

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect
import warnings

import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf import datetime642iso8601, time_clip
from .plot_spectr import plot_spectr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _time_avg(vdf, tint):
    if not tint:
        tint = list(datetime642iso8601(vdf.time.data[[0, -1]]))
        warnings.warn("Averages the entire time series", UserWarning)

    if len(tint) == 1:
        idx = bisect.bisect_left(vdf.time.data, np.datetime64(tint[0]))
        vdf_data = vdf.data.data[idx, ...]
        vdf_ener = vdf.energy.data[idx, ...]
        vdf_azim = vdf.phi.data[idx, ...]
        vdf_thet = vdf.theta.data

    elif len(tint) == 2:
        vdf = time_clip(vdf, tint)
        vdf_data = np.nanmean(vdf.data.data, axis=0)
        vdf_ener = np.nanmean(vdf.energy.data, axis=0)
        vdf_azim = np.nanmean(vdf.phi.data, axis=0)
        vdf_thet = vdf.theta.data
    else:
        raise TypeError("Invalid time interval format")

    vdf_new = xr.DataArray(
        vdf_data,
        coords=[vdf_ener, vdf_azim, vdf_thet],
        dims=["energy", "phi", "theta"],
    )

    return vdf_new


def _energy_avg(vdf, en_range):
    if not en_range:
        en_range = vdf.energy.data[[0, -1]]
        warnings.warn("Averages the entire energy range", UserWarning)
    else:
        en_range[0] = np.max(vdf.energy.data[0], en_range[0])
        en_range[1] = np.max(vdf.energy.data[-1], en_range[-1])

    idx = np.where(
        np.logical_and(
            vdf.energy.data > en_range[0],
            vdf.energy.data < en_range[1],
        ),
    )[0]
    assert idx, "Energy range is not covered by the instrument"

    out_data = np.nanmean(vdf.data[idx, ...], axis=0)

    out = xr.DataArray(
        out_data,
        coords=[vdf.phi.data, vdf.theta.data],
        dims=["phi", "theta"],
    )
    return out


def _check_units(vdf):
    if vdf.attrs["UNITS"] == "s^3/m^6":
        y_label = "PSD [s$^3$ m$^{-6}$]"
    elif vdf.attrs["UNITS"] == "1/(cm^2 s sr keV)":
        y_label = "Intensity [(cm$^2$ s sr keV)$^{-1}$]"
    else:
        raise ValueError("Invalid units")

    return y_label


def plot_ang_ang(vdf, tint: list = None, en_range: list = None):
    r"""Creates colormap of the phase space density or the differential
    particle flux, as a function of the azimuthal and elevation angles.

    Parameters
    ----------
    vdf : xarray.Dataset
        Skymap distribution.
    tint : list of str, Optional
        Time interval. If the time interval contains only one element,
        uses the distribution at the closest time. If the time interval
        contains two elements, time average the distribution. If None,
        uses the entire timeline for averaging. Default is None.
    en_range : list of float, Optional
        Energy range. If None uses the entire energy range.

    Returns
    -------
    f : matplotlib.figure.Figure
        Figure
    ax : matplotlib.axes._subplots.AxesSubplot
        Axis
    cax : matplotlib.axes._axes.Axes
        Colorbar axis

    """

    # Average over the selected time interval
    vdf_c = _time_avg(vdf, tint)

    # Average over the selected energy range
    vdf_avg = _energy_avg(vdf_c, en_range)

    f, ax = plt.subplots(1, figsize=(9, 9))
    f.subplots_adjust(left=0.1, right=0.85, bottom=0.1, top=0.9)
    ax, cax = plot_spectr(ax, vdf_avg, cscale="log")
    ax.set_xlabel("$\\phi$ [deg.]")
    ax.set_ylabel("$\\theta$ [deg.]")
    cax.set_ylabel(_check_units(vdf))
    ax.set_title(f"{en_range[0]:3.0f} keV < $E$ < {en_range[1]:3.0f} keV")

    return f, ax, cax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

# Built-in imports
import os

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..models import igrf
from .ts_tensor_xyz import ts_tensor_xyz
from .ts_vec_xyz import ts_vec_xyz
from .unix2datetime64 import unix2datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _triang(angle, axis):
    cos_angle = np.cos(np.deg2rad(angle))
    sin_angle = np.sin(np.deg2rad(angle))

    axes = list(filter(lambda x: x != axis, np.arange(3)))

    transf_mat = np.zeros((len(angle), 3, 3))
    transf_mat[:, axes[0], axes[0]] = cos_angle
    transf_mat[:, axes[1], axes[1]] = cos_angle
    transf_mat[:, axis, axis] = 1
    transf_mat[:, axes[0], axes[1]] = sin_angle
    transf_mat[:, axes[1], axes[0]] = -sin_angle

    return transf_mat


def _dipole_direction_gse(time, flag: str = "dipole"):
    lambda_, phi = igrf(time, flag)

    cos_phi = np.cos(np.deg2rad(phi))
    dipole_direction_geo_ = np.stack(
        [
            cos_phi * np.cos(np.deg2rad(lambda_)),
            cos_phi * np.sin(np.deg2rad(lambda_)),
            np.sin(np.deg2rad(phi)),
        ],
    ).T
    dipole_direction_gse_ = cotrans(
        ts_vec_xyz(unix2datetime64(time), dipole_direction_geo_),
        "geo>gse",
    )

    return dipole_direction_gse_


def _transformation_matrix(t, tind, hapgood, *args):
    t_zero, ut, d0_j2000, d_j2000, h_j2000, t_j2000 = args

    transf_mat_out = np.zeros((len(t), 3, 3))
    transf_mat_out[:, 0, 0] = np.ones(len(t))
    transf_mat_out[:, 1, 1] = np.ones(len(t))
    transf_mat_out[:, 2, 2] = np.ones(len(t))

    for j, t_num in enumerate(tind[::-1]):
        assert abs(t_num) in list(range(1, 6)), "t_num must be +/- 1, 2, 3, 4, 5"

        if t_num in [-1, 1]:
            if hapgood:
                theta = 100.461 + 36000.770 * t_zero + 15.04107 * ut

            else:
                # Source: United States Naval Observatory, Astronomical
                # Applications Dept. http: // aa.usno.navy.mil / faq / docs
                # / GAST.php Last modified: 2011/06/14T14:04
                gmst = 6.697374558
                gmst += 0.06570982441908 * d0_j2000
                gmst += 1.00273790935 * h_j2000
                gmst += 0.000026 * t_j2000**2

                gmst = gmst % 24  # Interval 0->24 hours
                theta = (360 / 24) * gmst  # Convert to degree.

            # invert if tInd = -1
            transf_mat = _triang(theta * np.sign(t_num), 2)

        elif t_num in [-2, 2]:
            if hapgood:
                eps = 23.439 - 0.013 * t_zero
                # Suns mean anomaly
                m_anom = 357.528 + 35999.050 * t_zero + 0.04107 * ut
                # Suns mean longitude
                m_long = 280.460 + 36000.772 * t_zero + 0.04107 * ut
                l_sun = m_long
                l_sun += (1.915 - 0.0048 * t_zero) * np.sin(
                    np.deg2rad(m_anom),
                )
                l_sun += 0.020 * np.sin(np.deg2rad(2 * m_anom))
            else:
                # Source: United States Naval Observatory, Astronomical
                # Applications Dept.
                # http://aa.usno.navy.mil/faq/docsSunApprox.php.
                # Last modified: 2012/11/06T14:12
                eps = 23.439 - 0.00000036 * d_j2000
                m_anom = 357.529 + 0.98560028 * d_j2000
                m_long = 280.459 + 0.98564736 * d_j2000
                l_sun = m_long
                l_sun += 1.915 * np.sin(np.deg2rad(m_anom))
                l_sun += 0.020 * np.sin(np.deg2rad(2 * m_anom))

            transf_mat = np.matmul(_triang(l_sun, 2), _triang(eps, 0))
            if t_num == -2:
                transf_mat = np.transpose(transf_mat, [0, 2, 1])

        elif t_num in [-3, 3]:
            dipole_direction_gse_ = _dipole_direction_gse(t, "dipole")
            y_e = dipole_direction_gse_[:, 1]  # 1st col is time
            z_e = dipole_direction_gse_[:, 2]
            psi = np.rad2deg(np.arctan(y_e / z_e))

            transf_mat = _triang(-psi * np.sign(t_num), 0)  # inverse if -3

        elif t_num in [-4, 4]:
            dipole_direction_gse_ = _dipole_direction_gse(t, "dipole")

            mu = np.arctan(
                dipole_direction_gse_[:, 0]
                / np.sqrt(np.sum(dipole_direction_gse_[:, 1:] ** 2, axis=1)),
            )
            mu = np.rad2deg(mu)

            transf_mat = _triang(-mu * np.sign(t_num), 1)

        else:
            lambda_, phi = igrf(t, "dipole")

            transf_mat = np.matmul(_triang(phi - 90, 1), _triang(lambda_, 2))
            if t_num == -5:
                transf_mat = np.transpose(transf_mat, [0, 2, 1])

        if j == 0:
            transf_mat_out = transf_mat
        else:
            transf_mat_out = np.matmul(transf_mat, transf_mat_out)

    return transf_mat_out


def cotrans(inp, flag, hapgood: bool = True):
    r"""Coordinate transformation GE0/GEI/GSE/GSM/SM/MAG as described in [1]_

    Parameters
    ----------
    inp : xarray.DataArray or ndarray
        Time series of the input field.
    flag : str
        Coordinates transformation "{coord1}>{coord2}", where coord1 and
        coord2 can be geo/gei/gse/gsm/sm/mag.
    hapgood : bool, Optional
        Indicator if original Hapgood sources should be used for angle
        computations or if updated USNO-AA sources should be used.
        Default = true, meaning original Hapgood sources.


    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu.pyrf import cotrans

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field in GSE coordinates

    >>> b_gse = get_data("b_gse_fgm_srvy_l2", tint, mms_id)

    Transform to GSM assuming that the original coordinates system is part
    of the inp metadata

    >>> b_gsm = cotrans(b_gse, 'GSM')

    If the original coordinates is not in the meta

    >>> b_gsm = cotrans(b_gse, 'GSE>GSM')

    Compute the dipole direction in GSE

    >>> dipole = cotrans(b_gse.time, 'dipoledirectiongse')


    References
    ----------
    .. [1]  Hapgood 1997 (corrected version of Hapgood 1992) Planet.Space Sci..Vol.
            40, No. 5. pp. 71l - 717, 1992

    """

    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"
    assert inp.ndim < 3, "inp must be scalar or vector"

    if ">" in flag:
        ref_syst_in, ref_syst_out = flag.split(">")
    else:
        ref_syst_in, ref_syst_out = [None, flag.lower()]

    if "COORDINATE_SYSTEM" in inp.attrs:
        ref_syst_internal = inp.attrs["COORDINATE_SYSTEM"].lower()
        ref_syst_internal = ref_syst_internal.split(">")[0]
    else:
        ref_syst_internal = None

    if ref_syst_in is not None and ref_syst_internal is not None:
        message = "input ref. frame in variable and input flag differs"
        assert ref_syst_internal == ref_syst_in, message
        flag = f"{ref_syst_in}>{ref_syst_out}"
    elif ref_syst_in is None and ref_syst_internal is not None:
        ref_syst_in = ref_syst_internal.lower()
        flag = f"{ref_syst_in}>{ref_syst_out}"
    elif flag.lower() == "dipoledirectiongse":
        flag = flag.lower()
    elif ref_syst_in is None and ref_syst_internal is None:
        raise ValueError(f"Transformation {flag} is unknown!")

    if ref_syst_in == ref_syst_out:
        return inp

    # J2000 reference time
    j2000 = 946727930.8160001
    # j2000 = Time("J2000", format="jyear_str").unix

    time = inp.time.data
    t = (time.astype(np.int64) * 1e-9).astype(np.float64)

    #  Terrestial Time (seconds since J2000)
    tts = t - j2000
    inp_ts = inp
    inp = inp.data

    if hapgood:
        day_start_epoch = time.astype("datetime64[D]")
        day_start_epoch = day_start_epoch.astype("datetime64[ns]")
        day_start_epoch = day_start_epoch.astype(np.int64) / 1e9
        mjd_ref_epoch = np.datetime64("2000-01-01T12:00:00", "ns")
        mjd_ref_epoch = mjd_ref_epoch.astype(np.int64) / 1e9

        # t_zero is time measured in Julian centuries from 2000-01-0112:00 UT
        # to the previous midnight
        t_zero = day_start_epoch - mjd_ref_epoch
        t_zero /= 3600 * 24 * 36525.0

        hours = (time.astype("datetime64[h]") - time.astype("datetime64[D]")).astype(
            float
        )
        minutes = (time.astype("datetime64[m]") - time.astype("datetime64[h]")).astype(
            float
        )
        seconds = 1e-9 * (
            time.astype("datetime64[ns]") - time.astype("datetime64[m]")
        ).astype(np.float64)
        ut = hours + minutes / 60 + seconds / 3600

        args_trans_mat = (t_zero, ut, None, None, None, None)

    else:
        # Julian date(of req.time) from J2000
        d_j2000 = tts / 86400

        # Julian date(of preceeding midnight of req.time) from J2000
        d0_j2000 = np.floor(tts / 86400) - 0.5

        # Julian centuries(of req.time) since J2000
        t_j2000 = d_j2000 / 36525

        # Hours in the of req.time(since midnight).
        h_j2000 = 24 * (d_j2000 - d0_j2000)

        args_trans_mat = (None, None, d0_j2000, d_j2000, h_j2000, t_j2000)

    if ">" in flag:
        root_path = os.path.dirname(os.path.abspath(__file__))
        file_name = "transformation_indices.json"

        with open(os.sep.join([root_path, file_name]), "r", encoding="utf-8") as file:
            transformation_dict = json.load(file)

        tind = transformation_dict[flag]

        transf_mat = _transformation_matrix(t, tind, hapgood, *args_trans_mat)

        if inp.ndim == 1:
            out = ts_tensor_xyz(inp_ts.time.data, transf_mat)

        else:
            out_data = np.einsum("kji,ki->kj", transf_mat, inp)
            out = inp_ts.copy()
            out.data = out_data
            out.attrs["COORDINATE_SYSTEM"] = ref_syst_out.upper()

    else:
        out = _dipole_direction_gse(t)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf import estimate
from .photo_current import photo_current
from .thermal_current import thermal_current

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["LangmuirProbe", "photo_current", "thermal_current"]


class LangmuirProbe:
    r"""Defines either spherical, cylindrical, conical or spherical +
    cylindrical/conical probes. Probe belonging to LangmuirProbe is defined
    with properties."""

    def __init__(
        self,
        name: str,
        surface: str = "cluster",
        r_sphere: Union[list, float, int] = None,
        r_wire: Union[list, float, int] = None,
        l_wire: Union[list, float, int] = None,
        s_photoemission: float = None,
    ):
        r"""Setup probe properties.

        Parameters
        ----------
        name : str
            Name of the probe.
        surface : str, Optional
            Surface materials or surface photoemission in [A/m^2].
            Default is "cluster".
        r_sphere : float, Optional
            Radius of the sphere in m.
        r_wire : float or list, Optional
            Radius of the wire in m. Can be a float for wire or a list of two
            floats for a stazer.
        l_wire : float, Optional
            Length of the wire/stazer in m.
        s_photoemission : float, Optional
            Surface photoemission in A/m^2. If not given obtain from surface
            type.

        Raises
        ------
        AssertionError : if r_wire as more than 2 elements.

        """

        message = (
            "The wire radius should be a numeric vector of "
            "length 1 (wire) or 2 (stazer)."
        )
        assert not r_wire or np.atleast_1d(r_wire) <= 2, message

        self.name = name
        self.wire = {"l": l_wire, "r": np.atleast_1d(r_wire)}
        self.sphere = {"r": r_sphere, "surface": surface}
        self.s_photoemission = s_photoemission

        # Probe type according to the specified parameters: Sphere + Wire or
        # Wire or Sphere
        self.get_probe_type()

        # Get probe area
        self.get_probe_area()

        # Get probe capacitance
        self.get_probe_capa()

    def get_probe_type(self):
        r"""Define probe type according to the specified parameters."""
        if self.sphere["r"] and self.wire["r"] and self.wire["l"]:
            self.probe_type = "sphere+wire"
        elif self.wire["r"] and self.wire["l"]:
            self.probe_type = "wire"
        elif self.sphere["r"]:
            self.probe_type = "sphere"
        else:
            self.probe_type = None

    def get_probe_area(self):
        r"""Computes probe area."""
        a_sphere_sunlit, a_sphere_total = [0, 0]
        a_wire_sunlit, a_wire_total = [0, 0]

        if isinstance(self.sphere["r"], (float, int)):
            a_sphere_sunlit = np.pi * self.sphere["r"] ** 2
            a_sphere_total = 4 * a_sphere_sunlit

        if (
            self.wire["r"]
            and isinstance(self.wire["r"], (float, int, list))
            and self.wire["l"]
            and isinstance(self.wire["l"], (float, int))
        ):
            a_wire_sunlit = 2 * np.mean(self.wire["r"]) * self.wire["l"]
            a_wire_total = np.pi * a_wire_sunlit

        self.area = {
            "sphere": a_sphere_total,
            "wire": a_wire_total,
            "total": a_sphere_total + a_wire_total,
            "sunlit": a_sphere_sunlit + a_wire_sunlit,
        }
        self.area["total_sunlit"] = self.area["total"] / self.area["sunlit"]
        self.area["sunlit_total"] = 1 / self.area["total_sunlit"]

    def get_probe_capa(self):
        r"""Computes probe capacitance.

        Raises
        ------
        ValueError : if length > radius.

        See Also
        --------
        pyrfu.pyrf.estimate.py

        """

        c_wire = 0
        c_sphere = estimate("capacitance_sphere", self.sphere["r"])

        if (
            self.wire["r"]
            and isinstance(self.wire["r"], (float, int, list))
            and self.wire["l"]
            and isinstance(self.wire["l"], (float, int))
        ):
            if self.wire["l"] > 10 * list([self.wire["r"]]):
                c_wire = estimate(
                    "capacitance_wire",
                    np.mean(self.wire["r"]),
                    self.wire["l"],
                )
            elif self.wire["l"] > list(self.wire["r"]):
                c_wire = estimate(
                    "capacitance_cylinder",
                    np.mean(self.wire["r"]),
                    self.wire["l"],
                )
            else:
                raise ValueError(
                    "estimate of capacitance for cylinder requires length > radius",
                )

        self.capacitance = np.sum([c_sphere, c_wire])

    def get_probe_surface_photoemission(self):
        r"""Computes (or get) surface photo emission."""
        if self.sphere["surface"]:
            self.s_photoemission = self.s_photoemission
        else:
            self.s_photoemission = photo_current(1.0, 0.0, 1.0, self.sphere["surface"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

g_corr = {
    "mms1-top6": 0.7,
    "mms1-top7": 2.5,
    "mms1-top8": 1.5,
    "mms1-bot5": 1.2,
    "mms1-bot6": 0.9,
    "mms1-bot7": 2.2,
    "mms1-bot8": 1.0,
    "mms2-top4": 1.2,
    "mms2-top6": 1.3,
    "mms2-top7": 0,
    "mms2-top8": 0.8,
    "mms2-bot6": 1.4,
    "mms2-bot7": 0,
    "mms2-bot8": 1.5,
    "mms3-top6": 0.7,
    "mms3-top7": 0.8,
    "mms3-top8": 1.0,
    "mms3-bot6": 0.9,
    "mms3-bot7": 0.9,
    "mms3-bot8": 1.3,
    "mms4-top6": 0.8,
    "mms4-top7": 0,
    "mms4-top8": 1.0,
    "mms4-bot6": 0.8,
    "mms4-bot7": 0.6,
    "mms4-bot8": 0.9,
    "mms4-bot9": 1.5,
}


def feeps_flat_field_corrections(inp_alle):
    r"""Apply flat field correction factors to FEEPS ion/electron
    data. Correct factors are from the gain factor found in:
    FlatFieldResults_V3.xlsx from Drew Turner, 1/19/2017

    Parameters
    ----------
    inp_alle : xarray.Dataset
        Dataset containing the energy spectrum of the available eyes of the
        Fly's Eye Energetic Particle Spectrometer (FEEPS).

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the available eyes of the
        Fly's Eye Energetic Particle Spectrometer (FEEPS) with corrected
        data.

    """

    # Spacecraft index
    mms_id = inp_alle.attrs["mmsId"]

    # List of sensors and eyes
    sensors_eyes = list(filter(lambda x: x[:3] in ["top", "bot"], inp_alle))

    out_dict = inp_alle.copy()

    for sensors_eye in sensors_eyes:
        sensor, eye = sensors_eye.split("-")
        correction = g_corr.get(f"mms{mms_id}-{sensor[:3]}{int(eye)}", 1.0)

        out_dict[sensors_eye].data *= correction

    out = xr.Dataset(out_dict)
    out.attrs = inp_alle.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import datetime

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def iso86012datetime(time):
    r"""Converts ISO 8601 time to datetime

    Parameters
    ----------
    time : ndarray or list or str
        Time

    Returns
    -------
    time_datetime : list
        Time in datetime format.

    """

    # Make sure that str is in ISO8601 format
    time = np.atleast_1d(time).astype("datetime64[ns]").astype(str)

    # ISO 8601 format with miliseconds precision (max precision for datetime)
    fmt = "%Y-%m-%dT%H:%M:%S.%f"

    # Convert to datetime format
    time_datetime = [datetime.datetime.strptime(t_[:-3], fmt) for t_ in time]

    return time_datetime



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import json

# Built-in imports
import os

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.iso86012datetime import iso86012datetime

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _bad_vars(bad_data):
    bad_vars_top = list(filter(lambda x: x not in [6, 7, 8], bad_data["top"]))
    bad_vars_bot = list(
        filter(lambda x: x not in [6, 7, 8], bad_data["bottom"]),
    )

    bad_vars = [
        *[f"top-{x}" for x in bad_vars_top],
        *[f"bottom-{x}" for x in bad_vars_bot],
    ]

    return bad_vars


def _bad_eyes(inp_dataset, bad_vars):
    inp_dataset_clean = inp_dataset.copy()

    for bad_var in bad_vars:
        if bad_var not in list(inp_dataset.keys()):
            continue

        inp_dataset_clean[bad_var].data[:] = np.nan

    return inp_dataset


def _bad_ch0(inp_dataset, bad_vars):
    inp_dataset_clean = inp_dataset.copy()

    for bad_var in bad_vars:
        if bad_var not in list(inp_dataset.keys()):
            continue
        # check if the energy table contains all nans
        energy = inp_dataset[inp_dataset[bad_var].dims[1]]
        if np.isnan(np.sum(energy.data)):
            continue

        inp_dataset_clean[bad_var].data[:, 0] = np.nan

    return inp_dataset_clean


def _bad_ch1(inp_dataset, bad_vars):
    inp_dataset_clean = inp_dataset.copy()

    for bad_var in bad_vars:
        if bad_var not in list(inp_dataset.keys()):
            continue
        # check if the energy table contains all nans
        energy = inp_dataset[inp_dataset[bad_var].dims[1]]
        if np.isnan(np.sum(energy.data)):
            continue

        inp_dataset_clean[bad_var].data[:, 0] = np.nan
        inp_dataset_clean[bad_var].data[:, 1] = np.nan

    return inp_dataset_clean


def _bad_ch2(inp_dataset, bad_vars):
    inp_dataset_clean = inp_dataset.copy()

    for bad_var in bad_vars:
        if bad_var not in list(inp_dataset.keys()):
            continue
        # check if the energy table contains all nans
        energy = inp_dataset[inp_dataset[bad_var].dims[1]]
        if np.isnan(np.sum(energy.data)):
            continue

        inp_dataset_clean[bad_var].data[:, 0] = np.nan
        inp_dataset_clean[bad_var].data[:, 1] = np.nan
        inp_dataset_clean[bad_var].data[:, 2] = np.nan

    return inp_dataset_clean


def feeps_remove_bad_data(inp_dataset):
    r"""This function removes bad eyes, bad lowest energy channels based on
    data from Drew Turner

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset with all active telescopes data.

    Returns
    -------
    inp_dataaset_clean_all : xarray.Dataset
        Dataset with all active telescopes data where bad eyes and lab lowest
        energy channels are set to NaN.

    """

    mms_id = inp_dataset.attrs["mmsId"]

    root_path = os.path.dirname(os.path.abspath(__file__))

    with open(
        os.path.join(root_path, "feeps_bad_data.json"), "r", encoding="utf-8"
    ) as file:
        feeps_bad_data = json.load(file)

    bad_data_table = feeps_bad_data["bad_data_table"]
    bad_ch0 = feeps_bad_data["bad_ch0"]
    bad_ch1 = feeps_bad_data["bad_ch1"]
    bad_ch2 = feeps_bad_data["bad_ch2"]

    # 1. BAD EYES
    # First, here is a list of the EYES that are bad, we need to make sure
    # these data are not usable (i.e., make all of the counts/rate/flux data
    # from these eyes NAN). These are for all modes, burst and survey:

    dates = [
        datetime.datetime.strptime(t_, "%Y-%m-%d").timestamp()
        for t_ in bad_data_table.keys()
    ]

    t_data = iso86012datetime(datetime642iso8601(inp_dataset.time.data[0]))[0]
    closest_table_tm = np.argmin([t_ - t_data.timestamp() for t_ in dates])

    closest_table = list(bad_data_table.keys())[closest_table_tm]
    bad_data = bad_data_table[closest_table][f"mms{mms_id:d}"]

    bad_vars_eyes = _bad_vars(bad_data)

    inp_dataset_clean_eye = _bad_eyes(inp_dataset, bad_vars_eyes)

    # 2. BAD LOWEST E-CHANNELS
    # Next, these eyes have bad first channels (i.e., lowest energy channel,
    # E-channel 0 in IDL indexing). Again, these data (just the
    # counts/rate/flux from the lowest energy channel ONLY!!!) should be
    # hardwired to be NAN for all modes (burst and both types of survey).
    # The eyes not listed here or above are ok though... so once we do this,
    # we can actually start showing the data down to the lowest levels (~33
    # keV), meaning we'll have to adjust the hard-coded ylim settings in
    # SPEDAS and the SITL software:

    if t_data > iso86012datetime(["2019-05-01T00:00:00.000"])[0]:
        bad_data_ch0 = bad_ch0[">2019-05-01"][f"mms{mms_id}"]
        bad_data_ch1 = bad_ch1[">2019-05-01"][f"mms{mms_id}"]
        bad_data_ch2 = bad_ch2[">2019-05-01"][f"mms{mms_id}"]
    else:
        bad_data_ch0 = bad_ch0["<2019-05-01"][f"mms{mms_id}"]
        bad_data_ch1 = bad_ch1["<2019-05-01"][f"mms{mms_id}"]
        bad_data_ch2 = bad_ch2["<2019-05-01"][f"mms{mms_id}"]

    bad_vars_ch0 = _bad_vars(bad_data_ch0)
    inp_dataset_clean_ch0 = _bad_ch0(inp_dataset_clean_eye, bad_vars_ch0)

    bad_vars_ch1 = _bad_vars(bad_data_ch1)
    inp_dataset_clean_ch1 = _bad_ch1(inp_dataset_clean_ch0, bad_vars_ch1)

    bad_vars_ch2 = _bad_vars(bad_data_ch2)
    inp_dataset_clean_all = _bad_ch2(inp_dataset_clean_ch1, bad_vars_ch2)

    return inp_dataset_clean_all



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

# Built-in imports
import os
import subprocess

# Local imports
from .list_files import list_files

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def copy_files(var, tint, mms_id, tar_path: str = "./data/"):
    r"""Copy files from local_data_dir as defined in config.json to the target
    path.

    Parameters
    ----------
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument.
            * var["tmmode"] : data rate.
            * var["lev"] : data level.
            * var["dtype"] : data type.
    tint : list of str
        Time interval.
    mms_id : str or int
        Index of the spacecraft.
    tar_path : str, Optional
        Target path. Default is './data/'.

    """

    # Normalize the target path and make sure it exists.
    tar_path = os.path.normpath(tar_path)
    assert os.path.exists(tar_path), f"{tar_path} doesn't exist!!"

    # Get the local_data_dir path from config.json.
    pkg_path = os.path.dirname(os.path.abspath(__file__))

    # Read the current version of the MMS configuration file
    with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
        config = json.load(fs)

    # Normalize the local_data_dir path and make sure it exists.
    mms_path = os.path.normpath(config["local_data_dir"]) + "/"
    assert os.path.exists(mms_path), f"{mms_path} doesn't exist!!"

    # List files that matches the requirements (instrument, date level,
    # data type, data rate) in the time interval for the target spacecraft.
    files = list_files(tint, mms_id, var)

    for file in files:
        relative_path = os.path.split(file)[0].replace(mms_path, "")
        path = os.path.join(tar_path, relative_path)
        target_file = os.path.join(path, os.path.split(file)[1])

        if not os.path.exists(path):
            os.makedirs(path)

        with subprocess.Popen(
            f"cp {file} {target_file}", stdout=subprocess.PIPE, shell=True
        ) as s_proc:
            (_, _) = s_proc.communicate()

            # This makes the wait possible
            _ = s_proc.wait()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# @Louis Richard
from .igrf import igrf
from .magnetopause_normal import magnetopause_normal

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["igrf", "magnetopause_normal"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools
import logging

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants

from ..pyrf.iso86012datetime64 import iso86012datetime64
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.ts_skymap import ts_skymap

# Local imports
from .psd_rebin import psd_rebin

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _coord_sys(coord_sys):
    x_vec = coord_sys[0, :] / np.linalg.norm(coord_sys[0, :])
    y_vec = coord_sys[1, :] / np.linalg.norm(coord_sys[1, :])

    z_vec = np.cross(x_vec, y_vec) / np.linalg.norm(np.cross(x_vec, y_vec))
    y_vec = np.cross(z_vec, x_vec) / np.linalg.norm(np.cross(z_vec, x_vec))

    changed_xyz = [False, False, False]

    for i, vec, comp in zip([0, 1, 2], [x_vec, y_vec, z_vec], ["x", "y", "z"]):
        if abs(np.rad2deg(np.arccos(np.dot(vec, coord_sys[:, i])))) > 1.0:
            logging.warning(
                "In making xyz a right handed orthogonal coordinate system, %(comp)s "
                "(in-plane %(i)d) was changed from %(x_old)s to %(x_new)s. Please "
                "verify that this is according to your intentions.",
                {
                    "comp": comp,
                    "i": i,
                    "x_old": np.array2string(coord_sys[:, i]),
                    "x_new": np.array2string(x_vec),
                },
            )
            changed_xyz[i] = True

    return x_vec, y_vec, z_vec, changed_xyz


def _init(vdf, tint):
    assert isinstance(vdf, xr.Dataset), "vdf must be a xarray.Dataset"

    len_e = 32

    if vdf.phi.data.ndim == 1:
        phi = xr.DataArray(
            np.tile(vdf.phi.data, (len(vdf.data), 1)),
            coords=[vdf.time.data, np.arange(len(vdf.phi.data))],
            dims=["time", "idx"],
        )
    else:
        phi = vdf.phi

    theta = vdf.theta.data
    polar = np.deg2rad(theta)
    azimuthal = np.deg2rad(phi)
    step_table = vdf.attrs.get("esteptable", np.zeros(len(vdf.time)))

    energy0 = vdf.attrs.get("energy0", vdf.energy.data[0, :])
    energy1 = vdf.attrs.get("energy1", vdf.energy.data[1, :])

    diff_energ = np.median(np.diff(np.log10(energy0))) / 2

    energy0_edges = np.hstack(
        [
            10 ** (np.log10(energy0) - diff_energ),
            10 ** (np.log10(energy0[-1]) + diff_energ),
        ],
    )
    energy1_edges = np.hstack(
        [
            10 ** (np.log10(energy1) - diff_energ),
            10 ** (np.log10(energy1[-1]) + diff_energ),
        ],
    )

    if tint is not None and len(tint) == 1:
        t_id = np.argmin(
            np.abs(vdf.time.data - iso86012datetime64(np.array(tint))[0]),
        )

        dist = vdf.data.data[t_id, ...]
        dist = dist[None, ...]
        step_table = step_table[t_id]
        azimuthal = azimuthal[t_id, ...].data

        if step_table.data:
            energy_edges = energy1_edges
        else:
            energy_edges = energy1_edges

    elif tint is not None and len(tint) == 2:
        dist = time_clip(vdf.data, tint)
        step_table = ts_scalar(vdf.time.data, step_table)
        step_table = time_clip(step_table, tint)
        azimuthal = time_clip(azimuthal, tint)

        if len(dist.time) > 1 and list(energy0) != list(energy1):
            logging.info("Rebinning distribution.")
            temp = ts_skymap(
                dist.time.data,
                dist.data,
                time_clip(vdf.energy, tint).data,
                np.rad2deg(azimuthal.data),
                theta,
            )
            newt, dist, energy, phi = psd_rebin(
                temp,
                phi,
                energy0,
                energy1,
                step_table,
            )
            dist = ts_skymap(
                newt,
                dist,
                np.tile(energy, (len(newt), 1)),
                phi,
                theta,
            )
            dist = time_clip(dist.data, tint)
            azimuthal = xr.DataArray(
                phi,
                coords=[newt, np.arange(phi.shape[1])],
                dims=["time", "odx"],
            )
            len_e = dist.shape[1]
            energy_edges = np.hstack(
                [
                    10 ** (np.log10(energy) - diff_energ / 2),
                    10 ** (np.log10(energy[-1]) + diff_energ / 2),
                ],
            )
        else:
            if all(step_table.data):
                energy_edges = energy1_edges
            else:
                energy_edges = energy0_edges

        dist = dist.data.data
        azimuthal = azimuthal.data
    else:
        raise ValueError("Invalid time interval")

    return dist, polar, azimuthal, energy_edges, len_e


def _cotrans(dist, polar, azimuthal, x_vec, y_vec, z_vec, e_lim, bin_corr):
    # Construct polar and azimuthal angle matrices
    polar = np.ones((len(dist), 1)) * polar

    f_mat = np.zeros((len(dist), dist.shape[2], dist.shape[1]))  #
    # azimuthal, energy
    edges_az = np.linspace(0, 2 * np.pi, azimuthal.shape[1] + 1)

    for i in range(len(dist)):
        pol_mat, azm_mat = np.meshgrid(polar[i, :], azimuthal[i, :])

        # '-' because the data shows which direction the particles were
        # coming from
        x_mat = -np.sin(pol_mat) * np.cos(azm_mat)
        y_mat = -np.sin(pol_mat) * np.sin(azm_mat)
        z_mat = -np.cos(pol_mat)

        # Transform into different coordinate system
        xx_mat = np.reshape(x_mat, (x_mat.shape[0] * x_mat.shape[1], 1))
        yy_mat = np.reshape(y_mat, (y_mat.shape[0] * y_mat.shape[1], 1))
        zz_mat = np.reshape(z_mat, (z_mat.shape[0] * z_mat.shape[1], 1))

        new_tmp_x = np.dot(np.hstack([xx_mat, yy_mat, zz_mat]), x_vec)
        new_tmp_y = np.dot(np.hstack([xx_mat, yy_mat, zz_mat]), y_vec)
        new_tmp_z = np.dot(np.hstack([xx_mat, yy_mat, zz_mat]), z_vec)

        new_x_mat = np.reshape(new_tmp_x, (x_mat.shape[0], x_mat.shape[1]))
        new_y_mat = np.reshape(new_tmp_y, (x_mat.shape[0], x_mat.shape[1]))
        new_z_mat = np.reshape(new_tmp_z, (x_mat.shape[0], x_mat.shape[1]))

        elevation_angle = np.arctan(
            new_z_mat / np.sqrt(new_x_mat**2 + new_y_mat**2),
        )
        plane_az = np.arctan2(new_y_mat, new_x_mat) + np.pi

        # gets velocity in direction normal to 'z'-axis
        geo_factor_elev = np.cos(elevation_angle)

        # geoFactorBinSize - detector bins in 'equator' plane are bigger and
        # get a larger weight. I think this is not good for the
        # implementation in this function
        if bin_corr:
            geo_factor_bin_size = np.sin(pol_mat)
        else:
            geo_factor_bin_size = np.ones(pol_mat.shape)

        f_mat[i, ...] = _cotrans_jit(
            dist[i, ...],
            elevation_angle,
            e_lim,
            plane_az,
            edges_az,
            geo_factor_elev,
            geo_factor_bin_size,
        )

    return f_mat


def _cotrans_jit(
    dist,
    elevation_angle,
    elevation_lim,
    plane_az,
    edges_az,
    geo_factor_elev,
    geo_factor_bin_size,
):
    out = np.zeros((dist.shape[1], dist.shape[0]))  # azimuthal, energy

    for i_en, i_az in itertools.product(
        range(dist.shape[0]),
        range(dist.shape[1]),
    ):
        # dist.data has dimensions nT x nE x nAz x nPol
        c_mat = dist[i_en, ...].copy()
        c_mat = c_mat * geo_factor_elev * geo_factor_bin_size
        c_mat[np.abs(elevation_angle) > np.deg2rad(elevation_lim)] = np.nan
        # use 0.1 deg to fix Az angle edges bug
        c_mat[plane_az < edges_az[i_az] - np.deg2rad(0.1)] = np.nan
        # use 0.1 deg to fix Az angle edges bug
        c_mat[plane_az > edges_az[i_az + 1] + np.deg2rad(0.1)] = np.nan

        out[i_az, i_en] = np.nanmean(c_mat)

    return out


def vdf_projection(
    vdf,
    tint,
    coord_sys: np.ndarray = np.eye(3),
    sc_pot: xr.DataArray = None,
    e_lim: float = 20,
    bins_correction: bool = False,
):
    r"""Computes projection of the velocity distribution onto a specified
    plane.

    Parameters
    ----------
    vdf : xarray.Dataset
        Electron or ion 3D skymap velocity distribution function.
    tint : list of str
        Computes data for time interval if len(tint) = 2 or closest time if
        len(tint) = 1. For tint includes two or more distributions the
        energies are rebinned into 64 channels.
    coord_sys : ndarray, Optional
        3x3 matrix with 1st column is x, 2nd column is y and 3rd column is z.
        z is normal to the projection plane and x and y are made orthogonal to
        z and each other if they are not already. Default is np.eye(3)
        (project onto spacecraft spin plane).
    sc_pot : xarray.DataArray, Optional
        Spacecraft potential to correct velocities. For a single value of tint
        the closest potential is used. For an interval the spacecraft
        potential is average over that interval. Default is None (no
        correction).
    e_lim : float, Optional
        Elevation angle limit in degrees above/below projection plane to
        include in projection. Default is e_lim = 20.
    bins_correction : bool, Optional
        Flag to correction elevation bins. Default is False.

    Returns
    -------
    v_x : ndarray
        2D grid of the velocities in the x direction.
    v_y : ndarray
        2D grid of the velocities in the y direction.
    f_mat : ndarray
        2D projection of the velocity distribution onto the specified plane

    """

    specie = vdf.attrs.get("species", "electrons")
    is_des = specie.lower() == "electrons"

    dist, polar, azimuthal, energy_edges, _ = _init(vdf, tint)
    x_vec, y_vec, z_vec, _ = _coord_sys(coord_sys)

    if azimuthal.ndim == 1:
        azimuthal = np.ones((len(dist), 1)) * azimuthal

    f_mat = _cotrans(
        dist,
        polar,
        azimuthal,
        x_vec,
        y_vec,
        z_vec,
        e_lim,
        bins_correction,
    )
    if len(dist) == 1:
        f_mat = np.squeeze(f_mat)
    else:
        f_mat = np.squeeze(np.nanmean(f_mat, axis=0))

    if sc_pot is not None:
        if len(tint) == 1:
            time_datetime64 = iso86012datetime64(np.array(tint))[0]
            t_id = np.argmin(np.abs(sc_pot.time.data - time_datetime64))
            sc_pot = sc_pot.data[t_id]
        else:
            sc_pot = time_clip(sc_pot, tint)
            sc_pot = np.nanmean(sc_pot.data)
    else:
        sc_pot = 0.0

    if is_des:
        mass = constants.electron_mass
    else:
        mass = constants.proton_mass
        sc_pot *= -1

    q_e = constants.elementary_charge

    speed_table = np.sqrt((energy_edges - sc_pot) * q_e * 2 / mass)
    speed_table = np.real(speed_table * 1e-3)  # km/s

    r_en = speed_table
    v_x = np.matmul(
        r_en[:, None],
        np.cos(np.linspace(0, 2 * np.pi, azimuthal.shape[1] + 1) + np.pi)[None, :],
    )
    v_y = np.matmul(
        r_en[:, None],
        np.sin(np.linspace(0, 2 * np.pi, azimuthal.shape[1] + 1) + np.pi)[None, :],
    )

    f_mat[f_mat <= 0] = np.nan

    return v_x, v_y, f_mat



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .calc_fs import calc_fs
from .cross import cross
from .dot import dot
from .normalize import normalize
from .resample import resample
from .time_clip import time_clip

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def poynting_flux(e_xyz, b_xyz, b_hat):
    r"""Estimates Poynting flux at electric field sampling as

    .. math::

        S = \frac{E \times B}{\mu_0}

    if `b0` is given project the Poynting flux along `b0`

    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.
    b_xyz : xarray.DataArray
        Time series of the magnetic field.
    b_hat : xarray.DataArray, Optional
        Time series of the direction to project the Pointing flux.

    Returns
    -------
    s : xarray.DataArray
        Time series of the Pointing flux.
    s_z : xarray.DataArray
        Time series of the projection of the Pointing flux (only if b0).
    int_s : xarray.DataArray
        Time series of the time integral of the Pointing flux
        (if b0 integral along b0).

    """

    # check which Poynting flux to calculate
    flag_s_z, flag_int_s_z, flag_int_s = [False, False, False]

    if b_hat is None:
        flag_int_s = True
    else:
        flag_s_z, flag_int_s_z = [True, True]

    # interval where both E & B exist
    tint = [
        np.max([np.min(e_xyz.time.data), np.min(b_xyz.time.data)]),
        np.min([np.max(e_xyz.time.data), np.max(b_xyz.time.data)]),
    ]
    tint = [np.datetime_as_string(time, "ns") for time in tint]

    e_xyz, b_xyz = [time_clip(e_xyz, tint), time_clip(b_xyz, tint)]

    if len(e_xyz) < len(b_xyz):
        e_xyz = resample(e_xyz, b_xyz)
        f_spl = calc_fs(b_xyz)
    elif len(e_xyz) > len(b_xyz):
        b_xyz = resample(b_xyz, e_xyz)
        f_spl = calc_fs(e_xyz)
    else:
        f_spl = calc_fs(b_xyz)

    # Calculate Poynting flux
    s_xyz = cross(e_xyz, b_xyz) / (4 * np.pi / 1e7) * 1e-9

    if flag_s_z:
        b_m = resample(b_hat, e_xyz)
        s_z = dot(normalize(b_m), s_xyz)
    else:
        s_z = None

    # time integral of Poynting flux along ambient magnetic field
    res = None

    if flag_int_s_z:
        s_z[np.isnan(s_z.data)] = 0  # set to zero points where Sz=NaN

        int_s_z = np.cumsum(s_z) / f_spl

        res = (s_xyz, s_z, int_s_z)

    if flag_int_s:  # time integral of all Poynting flux components
        # set to zero points where Sz=NaN
        s_xyz[np.isnan(s_xyz[:, 2].data)] = 0

        int_s = np.cumsum(s_xyz) / f_spl

        res = (s_xyz, int_s)

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib.pyplot as plt

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def colorbar(mappable, axis, pad: float = 0.01):
    r"""Add colorbar to ax corresponding to im.

    Parameters
    ----------
    mappable : matplotlib.collections.QuadMesh
        The quadrilateral mesh mappable described by this colorbar.
    axis : matplotlib.pyplot.subplotsaxes
        Axis of plot.
    pad : float, Optional
        Shift the colorbar with respect to the axis.

    Returns
    -------
    cax : matplotlib.colorbar.Colorbar
        Colorbar added to the plot.

    """

    pos = axis.get_position()
    fig = plt.gcf()
    cax = fig.add_axes([pos.x0 + pos.width + pad, pos.y0, pad, pos.height])
    fig.colorbar(mappable, cax)

    return cax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .cotrans import cotrans

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def l_shell(r_xyz):
    r"""Compute spacecraft position L Shell for a dipole magnetic field
    according to IRGF.

    Parameters
    ----------
    r_xyz : xarray.DataArray
        Time series of the spacecraft position. Must have a
        "COORDINATES_SYSTEM" attributes.

    Returns
    -------
    out : xarray.DataArray
        Time series of the spacecraft position L-Shell.

    """

    # Transform spacecraft coordinates to solar magnetic system
    r_sm = cotrans(r_xyz, "sm")

    # Compute Geomagnetic latitude
    lambda_ = np.arctan(r_sm[:, 2] / np.linalg.norm(r_sm[:, :2], axis=1))

    # Compute L shell
    out = np.linalg.norm(r_sm, axis=1) / np.cos(lambda_) ** 2

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def new_xyz(inp, trans_mat):
    r"""Transform the input field to the new frame.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field in the original coordinate system.
    trans_mat : array_like
        Transformation matrix.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input in the new frame.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)

    Compute MVA frame

    >>> b_lmn, l, mva = pyrf.mva(b_xyz)

    Move electric field to the MVA frame

    >>> e_lmn = pyrf.new_xyz(e_xyz, mva)

    """

    if inp.data.ndim == 3:
        out_data = np.matmul(np.matmul(trans_mat.T, inp.data), trans_mat)
    else:
        out_data = (trans_mat.T @ inp.data.T).T

    out = xr.DataArray(
        out_data,
        coords=inp.coords,
        dims=inp.dims,
        attrs=inp.attrs,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in import
import re
import warnings

# 3rd party imports
import numpy as np
import xarray as xr
from pycdfpp import DataType, load, to_datetime64

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.iso86012datetime64 import iso86012datetime64
from ..pyrf.time_clip import time_clip
from .get_variable import _pycdfpp_attributes_to_dict

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _shift_epochs(file, epoch):
    r"""Shift times for particles."""

    epoch_shifted = epoch["data"].copy()

    try:
        delta_minus_key = epoch["attrs"]["DELTA_MINUS_VAR"]
        delta_minus_var = {
            "data": file[delta_minus_key].values,
            "attrs": _pycdfpp_attributes_to_dict(file[delta_minus_key].attributes),
        }
        delta_plus_key = epoch["attrs"]["DELTA_PLUS_VAR"]
        delta_plus_var = {
            "data": file[delta_plus_key].values,
            "attrs": _pycdfpp_attributes_to_dict(file[delta_plus_key].attributes),
        }

        delta_vars = [delta_minus_var, delta_plus_var]
        flags_vars = [1e3, 1e3]  # Time scaling conversion flags

        for i, delta_var in enumerate(delta_vars):
            if isinstance(delta_var["attrs"], dict) and "UNITS" in delta_var["attrs"]:
                if delta_var["attrs"]["UNITS"].lower() == "s":
                    flags_vars[i] = 1e3
                elif delta_var["attrs"]["UNITS"].lower() == "ms":
                    flags_vars[i] = 1e0
                else:
                    message = " units are not clear, assume s"
                    warnings.warn(message)
            else:
                message = "Epoch_plus_var/Epoch_minus_var units are not clear, assume s"
                warnings.warn(message)

        flag_minus, flag_plus = flags_vars

        t_offset = (
            delta_plus_var["data"] * flag_plus - delta_minus_var["data"] * flag_minus
        )

        t_offset = (np.round(t_offset, 1) * 1e6 / 2).astype("timedelta64[ns]")
        t_diff = (
            delta_plus_var["data"] * flag_plus - delta_minus_var["data"] * flag_minus
        )
        t_diff = (np.round(t_diff, 1) * 1e6 / 2).astype("timedelta64[ns]")
        t_diff_data = np.median(np.diff(epoch["data"])) / 2

        if t_diff_data != np.mean(t_diff):
            t_offset = t_diff_data

        epoch_shifted += t_offset

        return {"data": epoch_shifted, "attrs": epoch["attrs"]}

    except KeyError:
        return {"data": epoch_shifted, "attrs": epoch["attrs"]}


def _get_epochs(file, cdf_name):
    r"""Get epochs form cdf and shift if needed."""

    depend0_key = file[cdf_name].attributes["DEPEND_0"][0]

    out = {
        "data": file[depend0_key].values,
    }

    if file[depend0_key].type == DataType.CDF_TIME_TT2000:
        try:
            out["data"] = to_datetime64(out["data"])

            # Get epoch attributes
            out["attrs"] = _pycdfpp_attributes_to_dict(file[depend0_key].attributes)

            # Shift times if particle data
            is_part = re.search(
                "^mms[1-4]_d[ei]s_",
                cdf_name,
            )  # Is it FPI data?
            is_part = is_part or re.search(
                "^mms[1-4]_hpca_",
                cdf_name,
            )  # Is it HPCA data?

            if is_part:
                out = _shift_epochs(file, out)

        except TypeError:
            pass

    return out


def _get_depend_attributes(file, depend_key):
    attributes = _pycdfpp_attributes_to_dict(file[depend_key].attributes)

    # Remove spaces in label
    try:
        attributes["LABLAXIS"] = attributes["LABLAXIS"].replace(" ", "_")

        if attributes["LABLAXIS"] == "Diffential_energy_channels":
            attributes["LABLAXIS"] = "Differential_energy_channels"

    except (KeyError, AttributeError):
        attributes["LABLAXIS"] = "comp"

    return attributes


def _get_depend(file, cdf_name, dep_num=1):
    out = {}

    if f"DEPEND_{dep_num:d}" in file[cdf_name].attributes:
        depend_key = file[cdf_name].attributes[f"DEPEND_{dep_num:d}"][0]
    elif f"REPRESENTATION_{dep_num:d}" in file[cdf_name].attributes:
        depend_key = file[cdf_name].attributes[f"REPRESENTATION_{dep_num:d}"][0]
    else:
        raise KeyError(f"no DEPEND_{dep_num:d}/REPRESENTATION_{dep_num:d} attributes")

    if depend_key == "x,y,z":
        out["data"] = np.array(depend_key.split(","))

        out["attrs"] = {"LABLAXIS": "comp"}
    else:
        out["data"] = file[depend_key].values

        if len(out["data"]) == 1:
            out["data"] = out["data"][0]

        if len(out["data"]) == 4 and all(
            out["data"].astype(str) == ["x", "y", "z", "r"]
        ):
            out["data"] = out["data"].astype(str)[:-1]

        elif out["data"].ndim == 2:
            if len(out["data"].flatten()) == 3:
                out["data"] = out["data"].flatten()
            else:
                try:
                    out["data"] = out["data"][0, :]
                except IndexError:
                    pass

        out["attrs"] = _get_depend_attributes(file, depend_key)

    return out


def get_ts(file_path, cdf_name, tint: list = None):
    r"""Reads field named cdf_name in file and convert to time series.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.
    cdf_name : str
        Name of the target variable in the cdf file.
    tint : list of str, Optional
        Time interval.

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable in the selected time interval.

    """

    # Check time interval type
    # Check time interval
    if tint is None:
        tint = ["1995-10-06T18:50:00.000000000", "2200-10-06T18:50:00.000000000"]
    elif isinstance(tint, (np.ndarray, list)):
        if isinstance(tint[0], np.datetime64):
            tint = datetime642iso8601(np.array(tint))
        elif isinstance(tint[0], str):
            tint = iso86012datetime64(
                np.array(tint),
            )  # to make sure it is ISO8601 ok!!
            tint = datetime642iso8601(np.array(tint))
        else:
            raise TypeError("Values must be in datetime64, or str!!")
    else:
        raise TypeError("tint must be array_like!!")

    out_dict = {}
    time, depend_1, depend_2, depend_3 = [{}, {}, {}, {}]

    # Load CDF file
    file = load(file_path)

    var_attrs = _pycdfpp_attributes_to_dict(file[cdf_name].attributes)
    glb_attrs = _pycdfpp_attributes_to_dict(file.attributes)
    out_dict["attrs"] = {"GLOBAL": glb_attrs, **var_attrs}
    out_dict["attrs"] = {k: out_dict["attrs"][k] for k in sorted(out_dict["attrs"])}

    assert "DEPEND_0" in var_attrs and "epoch" in var_attrs["DEPEND_0"].lower()

    time = _get_epochs(file, cdf_name)

    if time["data"] is None:
        return None

    if "DEPEND_1" in var_attrs or "REPRESENTATION_1" in var_attrs:
        depend_1 = _get_depend(file, cdf_name, 1)

    elif "afg" in cdf_name or "dfg" in cdf_name:
        depend_1 = {
            "data": ["x", "y", "z"],
            "attrs": {"LABLAXIS": "comp"},
        }

    if "DEPEND_2" in var_attrs or "REPRESENTATION_2" in var_attrs:
        depend_2 = _get_depend(file, cdf_name, 2)

        if depend_2["attrs"]["LABLAXIS"] == depend_1["attrs"]["LABLAXIS"]:
            depend_1["attrs"]["LABLAXIS"] = "rcomp"
            depend_2["attrs"]["LABLAXIS"] = "ccomp"

    if "DEPEND_3" in var_attrs or "REPRESENTATION_3" in var_attrs:
        if "REPRESENTATION_3" in var_attrs:
            assert out_dict["attrs"]["REPRESENTATION_3"] != "x,y,z"

        depend_3 = _get_depend(file, cdf_name, 3)

        if depend_3["attrs"]["LABLAXIS"] == depend_2["attrs"]["LABLAXIS"]:
            depend_2["attrs"]["LABLAXIS"] = "rcomp"
            depend_3["attrs"]["LABLAXIS"] = "ccomp"

    if "sector_mask" in cdf_name:
        cdf_name_mask = cdf_name.replace("sector_mask", "intensity")
        depend_1_key = file.varattsget(cdf_name_mask)["DEPEND_1"]

        depend_1["data"] = file[depend_1_key].values
        depend_1["attrs"] = _pycdfpp_attributes_to_dict(file[depend_1_key].attributes)

        depend_1["attrs"]["LABLAXIS"] = depend_1["attrs"]["LABLAXIS"].replace(" ", "_")

    if "edp_dce_sensor" in cdf_name:
        depend_1["data"] = ["x", "y", "z"]
        depend_1["attrs"] = {"LABLAXIS": "comp"}

    out_dict["data"] = file[cdf_name].values

    if out_dict["data"].ndim == 2 and out_dict["data"].shape[1] == 4:
        out_dict["data"] = out_dict["data"][:, :-1]
        # depend_1["data"] = depend_1["data"][:-1]

    if out_dict["data"].ndim == 2 and not depend_1:
        depend_1["data"] = np.arange(out_dict["data"].shape[1])
        depend_1["attrs"] = {"LABLAXIS": "idx"}

    if time and not depend_1 and not depend_2 and not depend_3:
        dims = ["time"]
        coords_data = [time["data"]]
        coords_attrs = [time["attrs"]]

    elif time and depend_1 and not depend_2 and not depend_3:
        dims = ["time", depend_1["attrs"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"]]
        coords_attrs = [time["attrs"], depend_1["attrs"]]

    elif time and depend_1 and depend_2 and not depend_3:
        dims = [
            "time",
            depend_1["attrs"]["LABLAXIS"],
            depend_2["attrs"]["LABLAXIS"],
        ]
        coords_data = [time["data"], depend_1["data"], depend_2["data"]]
        coords_attrs = [time["attrs"], depend_1["attrs"], depend_2["attrs"]]

    elif time and depend_1 and depend_2 and depend_3:
        dims = [
            "time",
            depend_1["attrs"]["LABLAXIS"],
            depend_2["attrs"]["LABLAXIS"],
            depend_3["attrs"]["LABLAXIS"],
        ]
        coords_data = [
            time["data"],
            depend_1["data"],
            depend_2["data"],
            depend_3["data"],
        ]
        coords_attrs = [
            time["attrs"],
            depend_1["attrs"],
            depend_2["attrs"],
            depend_3["attrs"],
        ]

    else:
        raise NotImplementedError

    out = xr.DataArray(
        out_dict["data"],
        coords=coords_data,
        dims=dims,
        attrs=out_dict["attrs"],
    )

    for dim, coord_attrs in zip(dims, coords_attrs):
        # Sort attributes and set to coordinates attribute
        out[dim].attrs = {k: coord_attrs[k] for k in sorted(coord_attrs)}

    # Time clip to original time interval
    out = time_clip(out, tint)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def solid_angle(inp0, inp1, inp2):
    r"""Calculates the solid angle of three vectors making up a triangle
    in a unit sphere with the sign taken into account.

    Parameters
    ----------
    inp0 : ndarray
        First vector.
    inp1 : ndarray
        Second vector.
    inp2 : ndarray
        Third vector.

    Returns
    -------
    angle : float
        Solid angle.

    """

    message = "must be array_like of xarray.DataArray"
    assert isinstance(inp0, (list, np.ndarray, xr.DataArray)), f"inp0 {message}"
    assert isinstance(inp1, (list, np.ndarray, xr.DataArray)), f"inp2 {message}"
    assert isinstance(inp2, (list, np.ndarray, xr.DataArray)), f"inp2 {message}"

    inp0, inp1, inp2 = [np.atleast_2d(inp) for inp in [inp0, inp1, inp2]]

    message = "must be a vector"
    assert inp0.shape[1] == 3, f"inp0 {message}"
    assert inp1.shape[1] == 3, f"inp1 {message}"
    assert inp2.shape[1] == 3, f"inp2 {message}"

    # Calculate the smaller angles between the vectors around origin
    acos_12 = np.arccos(np.sum(inp2 * inp1, axis=1))
    acos_02 = np.arccos(np.sum(inp0 * inp2, axis=1))
    acos_01 = np.arccos(np.sum(inp1 * inp0, axis=1))

    # Calculate the angles in the spherical triangle (Law of Cosines)
    alpha = np.arccos(
        (np.cos(acos_12) - np.cos(acos_02) * np.cos(acos_01))
        / (np.sin(acos_02) * np.sin(acos_01)),
    )
    beta = np.arccos(
        (np.cos(acos_02) - np.cos(acos_12) * np.cos(acos_01))
        / (np.sin(acos_12) * np.sin(acos_01)),
    )
    gamma = np.arccos(
        (np.cos(acos_01) - np.cos(acos_02) * np.cos(acos_12))
        / (np.sin(acos_02) * np.sin(acos_12)),
    )

    # Calculates the Surface area on the unit sphere (solid angle)
    angle = alpha + beta + gamma - np.pi
    # Calculate the sign of the area
    var = np.cross(inp2, inp1)
    div = np.sum(var * inp0, axis=1)
    sgn = np.sign(div)

    # Solid angle with sign taken into account
    angle = sgn * angle

    return angle



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib.pyplot as plt

from .add_position import add_position
from .annotate_heatmap import annotate_heatmap
from .colorbar import colorbar
from .make_labels import make_labels
from .mms_pl_config import mms_pl_config
from .pl_scatter_matrix import pl_scatter_matrix
from .pl_tx import pl_tx
from .plot_ang_ang import plot_ang_ang
from .plot_clines import plot_clines
from .plot_heatmap import plot_heatmap

# Local imports
from .plot_line import plot_line
from .plot_magnetosphere import plot_magnetosphere
from .plot_projection import plot_projection
from .plot_reduced_2d import plot_reduced_2d
from .plot_sitl_overview import plot_sitl_overview
from .plot_spectr import plot_spectr
from .plot_surf import plot_surf
from .span_tint import span_tint
from .zoom import zoom

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


# Setup plotting style
plt.style.use("classic")
plt.rcParams["figure.facecolor"] = "1"
plt.rcParams["mathtext.sf"] = "sans"
plt.rcParams["mathtext.fontset"] = "dejavusans"

__all__ = [
    "add_position",
    "annotate_heatmap",
    "colorbar",
    "make_labels",
    "mms_pl_config",
    "pl_scatter_matrix",
    "pl_tx",
    "plot_ang_ang",
    "plot_clines",
    "plot_heatmap",
    "plot_line",
    "plot_magnetosphere",
    "plot_projection",
    "plot_reduced_2d",
    "plot_sitl_overview",
    "plot_spectr",
    "plot_surf",
    "span_tint",
    "zoom",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_sector_spec(inp_alle):
    r"""Creates sector-spectrograms with FEEPS data (particle data organized
    by time and sector number)

    Parameters
    ----------
    inp_alle : xarray.Dataset
        Dataset of energy spectrum of all eyes.

    Returns
    -------
    out : xarray.Dataset
        Sector-spectrograms with FEEPS data for all eyes.

    """

    sensors_eyes_top = list(filter(lambda x: x[:3] in "top", inp_alle))
    sensors_eyes_bot = list(filter(lambda x: x[:3] in "bot", inp_alle))
    sensors_eyes = [*sensors_eyes_top, *sensors_eyes_bot]

    sector_time = inp_alle["spinsectnum"].time.data
    sector_data = inp_alle["spinsectnum"].data

    out_dict = {k: inp_alle[k] for k in inp_alle if k not in sensors_eyes}

    for sensors_eye in sensors_eyes:
        sensor_data = inp_alle[sensors_eye].data

        spin_starts = np.where(sector_data[:-1] > sector_data[1:])[0] + 1

        sector_spec = np.zeros((len(spin_starts), 64))

        c_start = spin_starts[0]

        for i, spin in enumerate(spin_starts):
            # find the sectors for this spin
            spin_sect = sector_data[c_start:spin]

            sector_spec[i, spin_sect] = np.nanmean(
                sensor_data[c_start:spin, :],
                axis=1,
            )

            c_start = spin

        out_dict[sensors_eye] = xr.DataArray(
            sector_spec,
            coords=[sector_time[spin_starts], np.arange(64)],
            dims=["time", "sectornum"],
        )

    out = xr.Dataset(out_dict)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import os

# Third party imports
import numpy as np
import requests
from scipy.io import readsav

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.extend_tint import extend_tint
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_time import ts_time

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

URL = "http://www.spedas.org/mms/mms_brst_intervals.sav"


def load_brst_segments(tint, data_path: str = None, download: bool = True):
    r"""Load burst segment time intervals associated with the input time
    interval `tint`.

    Parameters
    ----------
    tint : list
        Time interval to look for burst segments.
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu/mms/config.json`

    Returns
    -------
    brst_segments : list
        Segments of burst mode data.

    """

    # Check path
    if not data_path:
        pkg_path = os.path.dirname(os.path.abspath(__file__))

        # Read the current version of the MMS configuration file
        with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local_data_dir"])
    else:
        data_path = os.path.normpath(data_path)

    # Define path of the brst segment file to save
    file_path = os.path.join(data_path, URL.split("/", maxsplit=100)[-1])

    if download:
        # Get url content
        response = requests.get(URL, timeout=10)

        # Write content of the brst segment file to the local file
        with open(file_path, "wb", encoding="utf-8") as fs:
            fs.write(response.content)

    # Read brst segment content
    intervals = readsav(file_path)

    unix_start = ts_time(intervals["brst_intervals"].start_times[0])
    unix_end = ts_time(intervals["brst_intervals"].end_times[0])

    unix_start = time_clip(unix_start, extend_tint(tint, [-300, 300]))
    unix_end = time_clip(unix_end, extend_tint(tint, [-300, 300]))

    # +10 second offset added; there appears to be an extra 10
    # seconds of data, consistently, not included in the range here
    offset = np.timedelta64(10, "s")
    unix_end = unix_end.assign_coords(time=unix_end.time.data + offset)
    unix_end.data += np.timedelta64(10, "s")

    brst_segments = []
    l_bound, r_bound = [np.datetime64(t_) for t_ in tint]

    for start_time, end_time in zip(unix_start, unix_end):
        if end_time >= l_bound and start_time <= r_bound:
            segment = np.array([start_time.data, end_time.data])
            segment = list(datetime642iso8601(segment))
            brst_segments.append(segment)

    return brst_segments



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

# Local imports
from .calc_fs import calc_fs
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def avg_4sc(b_list):
    r"""Computes the input quantity at the center of mass of the MMS
    tetrahedron.

    Parameters
    ----------
    b_list : list of xarray.DataArray
        List of the time series of the quantity for each spacecraft.

    Returns
    -------
    b_avg : xarray.DataArray
        Time series of the input quantity a the enter of mass of the
        MMS tetrahedron.

    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu.pyrf import avg_4sc

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> b_mms = [get_data("B_gse_fgm_srvy_l2", tint, i) for i in range(1, 5)]
    >>> b_xyz = avg_4sc(b_mms)

    """

    # Check input type
    assert isinstance(b_list, list), "b_list must be a list"

    b_list_r = []

    for b in b_list:
        if isinstance(b, (xr.DataArray, xr.Dataset)):
            b_list_r.append(resample(b, b_list[0], f_s=calc_fs(b_list[0])))
        else:
            raise TypeError("elements of b_list must be DataArray or Dataset")

    # Average the resamples time series
    b_avg = sum(b_list_r) / len(b_list_r)

    return b_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def datetime642unix(time):
    r"""Converts datetime64 in ns units to unix time.

    Parameters
    ----------
    time : ndarray
        Time in datetime64 format.

    Returns
    -------
    time_unix : ndarray
        Time in unix format.

    See Also
    --------
    pyrfu.pyrf.unix2datetime64

    """

    # Make sure that time is in ns format
    if isinstance(time, np.datetime64):
        time = np.array([time])
        time_datetime64 = time.astype("datetime64[ns]")
    elif isinstance(time, (list, np.ndarray)) and isinstance(time[0], np.datetime64):
        time_datetime64 = time.astype("datetime64[ns]")
    else:
        raise TypeError("time must be numpy.datetime64 or array_like")

    # Convert to unix
    time_unix = time_datetime64.astype(np.int64) / 1e9

    return time_unix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import re
import warnings
from bisect import bisect_left
from datetime import datetime, timedelta

# 3rd party imports
import numpy as np
import requests
from dateutil.parser import parse

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

LASP_PUBL = "https://lasp.colorado.edu/mms/sdc/public/files/api/v1/"
LASP_SITL = "https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/"


def _login_lasp(user: str = "", password: str = ""):
    r"""Login to LASP colorado."""

    if not user:
        lasp_url = LASP_PUBL
    else:
        lasp_url = LASP_SITL

    session = requests.Session()
    session.auth = (user, password)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        _ = session.post("https://lasp.colorado.edu", verify=True, timeout=5)
        testget = session.get(lasp_url, verify=True, timeout=5)

    assert testget != "401", "Login failed!!"

    headers = {"User-Agent": "pyrfu"}

    return session, headers, lasp_url


def _construct_url_json_list(tint, mms_id, var, lasp_url):
    r"""Construct the url that return a json-formatted string of science
    filenames that are available for download according to:
    https://lasp.colorado.edu/mms/sdc/team/about/how-to/
    """

    tint = np.array(tint).astype("<M8[ns]").astype(str)
    tint = [datetime.strptime(t_[:-3], "%Y-%m-%dT%H:%M:%S.%f") for t_ in tint]
    start_date = tint[0].strftime("%Y-%m-%d")
    end_date = (tint[1] - timedelta(seconds=1)).strftime("%Y-%m-%d-%H-%M-%S")

    url = f"{lasp_url}/file_info/science"
    url = f"{url}?start_date={start_date}&end_date={end_date}&sc_id=mms{mms_id}"

    url = f"{url}&instrument_id={var['inst']}"
    url = f"{url}&data_rate_mode={var['tmmode']}"
    url = f"{url}&data_level={var['lev']}"

    if var["dtype"]:
        url = f"{url}&descriptor={var['dtype']}"

    return url


def _files_in_interval(in_files, trange):
    r"""Filters the file list returned by the SDC to the requested time
    range. This filter is purposefully liberal, it regularly grabs an extra
    file due to special cases
    """

    file_name = r"mms.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    file_times = []

    regex = re.compile(file_name)

    for file in in_files:
        matches = regex.match(file["file_name"])
        if matches:
            file_times.append(
                (
                    file["file_name"],
                    parse(matches.groups()[0]).timestamp(),
                    file["timetag"],
                    file["file_size"],
                ),
            )

    # sort in time
    sorted_files = sorted(file_times, key=lambda x: x[1])

    times = [t[1] for t in sorted_files]

    idx_min = bisect_left(times, parse(trange[0]).timestamp())

    def mkout(f):
        return {"file_name": f[0], "timetag": f[2], "size": f[3]}

    if idx_min == 0:
        files = list(map(mkout, sorted_files[idx_min:]))
    else:
        files = list(map(mkout, sorted_files[idx_min - 1 :]))

    return files


def _make_urls_cdfs(lasp_url, files):
    for file in files:
        file["url"] = f"{lasp_url}download/science?file={file['file_name']}"

    return files


def list_files_sdc(tint, mms_id, var, login: str = "", password: str = ""):
    r"""Find availables files from the LASP SDC for the target instrument,
    data type, data rate, mms_id and level during the target time interval.

    Parameters
    ----------
    tint : array_like
        Time interval
    mms_id : str or int
        Index of the spacecraft
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument
            * var["tmmode"] : data rate
            * var["lev"] : data level
            * var["dtype"] : data type
    login : str, Optional
        Login to LASP MMS SITL. Default downloads from
        https://lasp.colorado.edu/mms/sdc/public/
    password : str, Optional
        Password to LASP MMS SITL. Default downloads from
        https://lasp.colorado.edu/mms/sdc/public/

    Returns
    -------
    file_names : list
        List of files corresponding to the parameters in the selected time
        interval

    """
    sdc_session, headers, lasp_url = _login_lasp(login, password)

    url_json_cdfs = _construct_url_json_list(tint, mms_id, var, lasp_url)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        http_json = sdc_session.get(url_json_cdfs, verify=True, headers=headers).json()

    file_names = _files_in_interval(http_json["files"], tint)
    sdc_session.close()

    file_names = _make_urls_cdfs(lasp_url, file_names)

    return file_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# Third party imports
import numpy as np
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["shock_parameters"]

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def shock_parameters(spec):
    r"""Calculate shock related plasma parameters.

    Parameters
    ----------
    spec : dict
        Hash table with parameters input with fixed names. After the parameter name,
        a name of the region can be given, e.g. "u" and "d". All parameters except b
        are optional.

    Returns
    -------
    out : dict
        Hash table with derived plasma parameters from hash table spec with measured
        plasma parameters.

    """

    id_b = list(filter(lambda x: x[0].lower() == "b", spec))
    regions = [id_[1:] for id_ in id_b if len(id_) > 1]
    regions = regions if len(regions) > 1 else [""]

    spec["ref_sys"] = spec.get("ref_sys", "sc")
    assert spec["ref_sys"].lower() in ["sc", "nif"], "Invalid reference frame"

    if spec["ref_sys"].lower() == "nif":
        if "nvec" not in spec:
            logging.warning("Setting shock speed, nvec to [1, 0, 0]")
            spec["nvec"] = np.array([1, 0, 0])

        if "v_sh" not in spec:
            logging.warning("Setting shock speed, Vsh, to 0.")
            spec["v_sh"] = 0.0

    dspec = {}

    # Alfven speed
    if f"n{regions[0]}" in spec:
        for region in regions:
            dspec[f"v_a{region}"] = _v_alfv(spec[f"b{region}"], spec[f"n{region}"])

    # Sound speed
    if f"t_i{regions[0]}" in spec and f"t_e{regions[0]}" in spec:
        for region in regions:
            dspec[f"v_ts{region}"] = _v_sound(
                spec[f"t_i{region}"], spec[f"t_e{region}"]
            )

    # Fast speed
    if (
        f"n{regions[0]}" in spec
        and f"v{regions[0]}" in spec
        and f"t_i{regions[0]}" in spec
        and f"t_e{regions[0]}" in spec
    ):
        for region in regions:
            dspec[f"v_f{region}"] = _v_fast(
                spec[f"b{region}"],
                spec[f"n{region}"],
                spec[f"v{region}"],
                spec[f"t_i{region}"],
                spec[f"t_e{region}"],
            )

    # Normal incidence frame velocity
    if f"v{regions[0]}" in spec and "nvec" in spec and "v_sh" in spec:
        for region in regions:
            dspec[f"v_nif{region}"] = _nif_speed(
                spec[f"v{region}"], spec["v_sh"], spec["nvec"]
            )

    # de Hoffman-Teller frame velocity
    if f"v{regions[0]}" in spec and "nvec" in spec and "v_sh" in spec:
        for region in regions:
            dspec[f"v_htf{region}"] = _htf_speed(
                spec[f"b{region}"], spec[f"v{region}"], spec["v_sh"], spec["nvec"]
            )

    # Ion gyrofrequency
    for region in regions:
        dspec[f"f_cp{region}"] = _ion_gyro_freq(spec[f"b{region}"])

    # Ion inertial length
    if f"n{regions[0]}" in spec:
        for region in regions:
            dspec[f"l_i{region}"] = _ion_in_len(spec[f"n{region}"])

    if f"n{regions[0]}" in spec:
        for region in regions:
            dspec[f"r_cp{region}"] = _ion_gyro_rad(
                spec[f"b{region}"], spec[f"v{region}"]
            )

    # Alfven Mach number
    if f"n{regions[0]}" in spec and f"v{regions[0]}" in spec:
        for region in regions:
            dspec[f"m_a{region}"] = _alfv_mach(
                spec[f"b{region}"],
                spec[f"n{region}"],
                spec[f"v{region}"],
                spec["ref_sys"],
                spec["v_sh"],
                spec["nvec"],
            )

    # Sonic Mach number
    if (
        f"v{regions[0]}" in spec
        and f"t_i{regions[0]}" in spec
        and f"t_e{regions[0]}" in spec
    ):
        for region in regions:
            dspec[f"m_s{region}"] = _sonic_mach(
                spec[f"v{region}"],
                spec[f"t_i{region}"],
                spec[f"t_e{region}"],
                spec["ref_sys"],
                spec["v_sh"],
                spec["nvec"],
            )

    if (
        f"n{regions[0]}" in spec
        and f"v{regions[0]}" in spec
        and f"t_i{regions[0]}" in spec
        and f"t_e{regions[0]}" in spec
    ):
        for region in regions:
            dspec[f"m_f{region}"] = _fast_mach(
                spec[f"b{region}"],
                spec[f"n{region}"],
                spec[f"v{region}"],
                spec[f"t_i{region}"],
                spec[f"t_e{region}"],
                spec["ref_sys"],
                spec["v_sh"],
                spec["nvec"],
            )

    if f"n{regions[0]}" in spec and f"t_i{regions[0]}" in spec:
        for region in regions:
            dspec[f"beta_i{region}"] = _beta(
                spec[f"b{region}"], spec[f"n{region}"], spec[f"t_i{region}"]
            )

    if f"n{regions[0]}" in spec and f"t_e{regions[0]}" in spec:
        for region in regions:
            dspec[f"beta_e{region}"] = _beta(
                spec[f"b{region}"], spec[f"n{region}"], spec[f"t_e{region}"]
            )

    return dspec


def _ion_gyro_freq(b):
    b_si = 1e-9 * np.linalg.norm(b)
    w_cp = constants.elementary_charge * b_si / constants.proton_mass
    return w_cp / (2 * np.pi)


def _ion_in_len(n):
    n_si = 1e6 * n
    w_pp = np.sqrt(
        n_si
        * constants.elementary_charge**2
        / (constants.proton_mass * constants.epsilon_0)
    )
    l_i = constants.speed_of_light / w_pp
    return l_i


def _ion_gyro_rad(b, v):
    b_si = 1e-9 * np.linalg.norm(b)
    e_i = 0.5 * constants.proton_mass * np.linalg.norm(v) ** 2 / constants.electron_volt
    v_tp = constants.speed_of_light * np.sqrt(
        1
        - 1
        / (
            e_i
            * constants.elementary_charge
            / (constants.proton_mass * constants.speed_of_light**2)
            + 1
        )
        ** 2
    )
    gamma_p = 1 / np.sqrt(1 - (v_tp / constants.speed_of_light) ** 2)
    rho_p = (
        constants.proton_mass
        * constants.speed_of_light
        / (constants.elementary_charge * b_si)
        * np.sqrt(gamma_p**2 - 1)
    )
    return rho_p


def _v_alfv(b, n):
    b_si = 1e-9 * np.linalg.norm(b)
    n_si = 1e6 * n
    return b_si / np.sqrt(constants.mu_0 * n_si * constants.proton_mass)


def _v_sound(t_i, t_e):
    t_i_si = t_i * constants.electron_volt
    t_e_si = t_e * constants.electron_volt
    return np.sqrt((t_e_si + 3 * t_i_si) / constants.proton_mass)


def _v_fast(b, n, v, t_i, t_e, theta=None):
    if theta is None:
        theta = np.arccos(np.sum(b * v) / (np.linalg.norm(b) * np.linalg.norm(v)))
    else:
        theta = np.deg2rad(theta)

    v_a = _v_alfv(b, n)
    c_s = _v_sound(t_i, t_e)
    c_ms0 = np.sqrt(v_a**2 + c_s**2)
    v_f = np.sqrt(
        c_ms0**2 / 2
        + np.sqrt(c_ms0**4 / 4 - v_a**2.0 * c_s**2 * np.cos(theta) ** 2)
    )
    return v_f


def _nif_speed(v, v_sh, nvec):
    v_si = 1e3 * v
    v_sh_si = 1e3 * v_sh
    v_nif = v_si - (np.sum(v_si * nvec) - v_sh_si) * nvec
    return v_nif


def _htf_speed(b, v, v_sh, nvec):
    v_si = 1e3 * v
    v_sh_si = 1e3 * v_sh

    # first get the velocity in a shock rest frame
    v_in_shock_rest_frame = v_si - v_sh_si * nvec

    # then get the dHT frame speed in the shock rest frame
    v_htf_srf = np.cross(nvec, np.cross(v_in_shock_rest_frame, b)) / np.sum(b * nvec)

    # then the dHT frame speed in the sc frame is the shock speed plus  the dHT frame
    # speed in the shock rest frame (I think)
    v_htf = v_sh_si * nvec + v_htf_srf
    return v_htf


def _alfv_mach(b, n, v, ref_sys, v_sh, nvec):
    v_si = 1e3 * v
    v_sh_si = 1e3 * v_sh

    if ref_sys.lower() == "nif" and nvec is not None:
        m_a = np.abs(np.sum(v_si * nvec) - v_sh_si) / _v_alfv(b, n)
    else:
        m_a = np.linalg.norm(v_si) / _v_alfv(b, n)

    return m_a


def _sonic_mach(v, t_i, t_e, ref_sys, v_sh, nvec):
    v_si = 1e3 * v
    v_sh_si = 1e3 * v_sh

    if ref_sys.lower() == "nif" and nvec is not None:
        m_s = np.abs(np.sum(v_si * nvec) - v_sh_si) / _v_sound(t_i, t_e)
    else:
        m_s = np.linalg.norm(v_si) / _v_sound(t_i, t_e)

    return m_s


def _fast_mach(b, n, v, t_i, t_e, ref_sys, v_sh, nvec):
    v_si = 1e3 * v
    v_sh_si = 1e3 * v_sh

    if ref_sys.lower() == "nif" and nvec is not None:
        theta_bn = np.rad2deg(np.arccos(np.sum(b * nvec) / np.linalg.norm(b)))
        m_f = np.abs(np.sum(v_si * nvec) - v_sh_si) / _v_fast(
            b, n, v, t_i, t_e, theta_bn
        )
    else:
        m_f = np.linalg.norm(v_si) / _v_fast(b, n, v, t_i, t_e)

    return m_f


def _beta(b, n, t_s):
    b_si = 1e-9 * np.linalg.norm(b)
    n_si = 1e6 * n
    t_s_si = t_s * constants.electron_volt
    beta_s = (n_si * t_s_si) / (b_si**2 / (2 * constants.mu_0))
    return beta_s



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _mass_ratio(inp):
    if inp.attrs["species"].lower() in ["ions", "ion", "protons", "proton"]:
        mass_ratio = 1
    elif inp.attrs["species"].lower() in ["alphas", "alpha", "helium"]:
        mass_ratio = 4
    elif inp.attrs["species"].lower() in ["electrons", "e"]:
        mass_ratio = constants.electron_mass / constants.proton_mass
    else:
        raise ValueError("Invalid specie")

    return mass_ratio


def _convert(inp, units, mass_ratio):
    fact = 1e6 * 0.53707 * mass_ratio**2

    if units.lower() == "1/(cm^2 s sr kev)":
        tmp_data = inp * 1e-3 / 1e18 * fact
    elif units.lower() == "1/(cm^2 s sr ev)":
        tmp_data = inp / 1e18 * fact
    else:
        raise ValueError("Invalid unit")

    return tmp_data


def dpf2psd(inp):
    r"""Computes phase space density from differential particle flux.

    Parameters
    ----------
    inp : xarray.Dataset or xarray.DataArray
        Time series of the differential particle flux in
        [(cm^{2} s sr keV)^{-1}].

    Returns
    -------
    psd : xarray.Dataset
        Time series of the phase space density in [s^{3} m^{-6}].

    """

    assert isinstance(inp, (xr.DataArray, xr.Dataset)), "inp must be a xarray"

    if isinstance(inp, xr.Dataset):
        tmp_data = _convert(inp.data.data, inp.data.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy[:, :, None, None], (1, 1, *tmp_data.shape[2:]))
        tmp_data /= energy_mat
        out = inp.copy()
        out.data.data = np.squeeze(tmp_data)
        out.data.attrs["UNITS"] = "s^3/m^6"
    else:
        tmp_data = _convert(inp.data, inp.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy, (tmp_data.shape[0], 1))
        tmp_data /= energy_mat
        out = inp.copy()
        out.data = np.squeeze(tmp_data)
        out.attrs["UNITS"] = "s^3/m^6"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def histogram2d(inp1, inp2, bins=100, y_range=None, weights=None, density=True):
    r"""Computes 2d histogram of inp2 vs inp1 with nbins number of bins.

    Parameters
    ----------
    inp1 : xarray.DataArray
        Time series of the x coordinates of the points to be histogrammed.
    inp2 : xarray.DataArray
        Time series of the y coordinates of the points to be histogrammed.
    bins : str or int or tuple, Optional
        Number of bins. Default is ``bins=100``.
    y_range : array_like, shape(2,2), Optional
        The leftmost and rightmost edges of the bins along each dimension
        (if not specified explicitly in the `bins` parameters):
        ``[[xmin, xmax], [ymin, ymax]]``. All values outside of this range
        will be considered outliers and not tallied in the histogram.
    weights : array_like, shape(N,), Optional
        An array of values ``w_i`` weighing each sample ``(x_i, y_i)``.
        Weights are normalized to 1 if `normed` is True. If `normed` is
        False, the values of the returned histogram are equal to the sum of
        the weights belonging to the samples falling into each bin.
    density : bool, Optional
        If False, the default, returns the number of samples in each bin.
        If True, returns the probability *density* function at the bin,
        ``bin_count / sample_count / bin_area``.

    Returns
    -------
    out : xarray.DataArray
        2D map of the density of ``inp2`` vs ``inp1``.

    Examples
    --------
    >>> import numpy as np
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_id = np.arange(1, 5)

    Load magnetic field and electric field

    >>> r_mms = [mms.get_data("r_gse", tint, i) for i in mms_id]
    >>> b_mms = [mms.get_data("b_gse_fgm_srvy_l2", tint, i) for i in mms_id]

    Compute current density, etc

    >>> j_xyz, _, b_xyz, _, _, _ = pyrf.c_4_j(r_mms, b_mms)

    Compute magnitude of B and J

    >>> b_mag = pyrf.norm(b_xyz)
    >>> j_mag = pyrf.norm(j_xyz)

    Histogram of J vs B

    >>> h2d_b_j = pyrf.histogram2d(b_mag, j_mag)

    """

    # resample inp2 with respect to inp1
    if len(inp2) != len(inp1):
        inp2 = resample(inp2, inp1)

    h2d, x_edges, y_edges = np.histogram2d(
        inp1.data,
        inp2.data,
        bins=bins,
        range=y_range,
        density=density,
        weights=weights,
    )

    x_bins = x_edges[:-1] + np.median(np.diff(x_edges)) / 2
    y_bins = y_edges[:-1] + np.median(np.diff(y_edges)) / 2

    out = xr.DataArray(
        h2d,
        coords=[x_bins, y_bins],
        dims=["x_bins", "y_bins"],
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
from matplotlib.transforms import TransformedBbox, blended_transform_factory
from mpl_toolkits.axes_grid1.inset_locator import (
    BboxConnector,
    BboxConnectorPatch,
    BboxPatch,
)

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _connect_bbox(
    bbox1,
    bbox2,
    loc1a,
    loc2a,
    loc1b,
    loc2b,
    prop_lines,
    prop_patches: dict = None,
):
    if prop_patches is None:
        prop_patches = {**prop_lines, "alpha": prop_lines.get("alpha", 1) * 0}

    connector_a = BboxConnector(
        bbox1,
        bbox2,
        loc1=loc1a,
        loc2=loc2a,
        **prop_lines,
    )
    connector_a.set_clip_on(False)
    connector_b = BboxConnector(
        bbox1,
        bbox2,
        loc1=loc1b,
        loc2=loc2b,
        **prop_lines,
    )
    connector_b.set_clip_on(False)

    bbox_patch1 = BboxPatch(bbox1, **prop_patches)
    bbox_patch2 = BboxPatch(bbox2, **prop_patches)

    connector_patch = BboxConnectorPatch(
        bbox1,
        bbox2,
        loc1a=loc1a,
        loc2a=loc2a,
        loc1b=loc1b,
        loc2b=loc2b,
        **prop_patches,
    )

    connector_patch.set_clip_on(False)

    return connector_a, connector_b, bbox_patch1, bbox_patch2, connector_patch


def zoom(ax1, ax2, **kwargs):
    r"""Similar to zoom_effect01.  The xmin & xmax will be taken from the
    ax1.viewLim.

    Parameters
    ----------
    ax1 : matplotlib.pyplot.subplotsaxes
        Reference axes.
    ax2 : matplotlib.pyplot.subplotsaxes
        Connected axes.

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments control the BboxPatch.

    Returns
    -------
    ax1 : matplotlib.pyplot.subplotsaxes
        Reference axis.
    ax2 : matplotlib.pyplot.subplotsaxes
        Connected axis.

    """

    tt = ax1.transScale + (ax1.transLimits + ax2.transAxes)
    trans = blended_transform_factory(ax2.transData, tt)

    bbox_1 = ax1.bbox
    bbox_2 = TransformedBbox(ax1.viewLim, trans)

    c1, c2, p1, p2, p = _connect_bbox(
        bbox_1,
        bbox_2,
        loc1a=2,
        loc2a=3,
        loc1b=1,
        loc2b=4,
        prop_lines=kwargs,
    )

    ax1.add_patch(p1)
    ax2.add_patch(p2)
    ax2.add_patch(c1)
    ax2.add_patch(c2)
    ax2.add_patch(p)

    return ax1, ax2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def spectr_to_dataset(spectr):
    r"""Convert energy spectrum (DataArray) to Dataset.

    Parameters
    ----------
    spectr : xarray.DataArray
        Spectrogram in DataArray format.

    Returns
    -------
    out : xarray.Dataset
        Spectrogram in Dataset format.

    """

    time = spectr.time.data
    energy = spectr.energy.data
    energy = np.tile(energy, (len(time), 1))

    data = spectr.data

    out_dict = {
        "data": (["time", "idx0"], data),
        "energy": (["time", "idx0"], energy),
        "time": time,
        "idx0": np.arange(energy.shape[1]),
    }

    out = xr.Dataset(out_dict)

    out.attrs = spectr.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

from .calc_fs import calc_fs

# Local imports
from .resample import resample
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def convert_fac(inp, b_bgd, r_xyz: list = None):
    r"""Transforms to a field-aligned coordinate (FAC) system defined as :
        * R_parallel_z aligned with the background magnetic field
        * R_perp_y defined by R_parallel cross the position vector of the
            spacecraft (nominally eastward at the equator)
        * R_perp_x defined by R_perp_y cross R_par

    If inp is one vector along r direction, out is inp[perp, para]
    projection.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field.
    b_bgd : xarray.DataArray
        Time series of the background magnetic field.
    r_xyz : xarray.DataArray or ndarray or list
        Position vector of spacecraft.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input field in field aligned coordinates
        system.

    Notes
    -----
    All input parameters must be in the same coordinate system.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field (FGM) and electric field (EDP)

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)

    Convert to field aligned coordinates

    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, numpy.array([1, 0, 0]))

    """

    # Check input type
    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"
    assert isinstance(b_bgd, xr.DataArray), "b_xyz must be a xarray.DataArray"

    assert r_xyz is None or isinstance(
        r_xyz,
        (xr.DataArray, list, np.ndarray),
    )

    if len(inp) != len(b_bgd):
        b_bgd = resample(b_bgd, inp, f_s=calc_fs(inp))

    time, inp_data = [inp.time.data, inp.data]

    if r_xyz is None:
        r_xyz = np.array([1, 0, 0])

    if isinstance(r_xyz, xr.DataArray):
        r_xyz = resample(r_xyz, b_bgd, f_s=calc_fs(b_bgd))

    elif isinstance(r_xyz, (list, np.ndarray)) and len(r_xyz) == 3:
        r_xyz = np.tile(r_xyz, (len(b_bgd), 1))

    if b_bgd.ndim == 2 and b_bgd.shape[1] == 3:
        # Normalize background magnetic field
        b_hat = b_bgd / np.linalg.norm(b_bgd, axis=1, keepdims=True)

        # Perpendicular
        r_perp_y = np.cross(b_hat, r_xyz, axis=1)
        r_perp_y /= np.linalg.norm(r_perp_y, axis=1, keepdims=True)
        r_perp_x = np.cross(r_perp_y, b_bgd, axis=1)
        r_perp_x /= np.linalg.norm(r_perp_x, axis=1, keepdims=True)
        r_para = b_hat
    elif b_bgd.ndim == 3 and b_bgd.shape[1] == 3 and b_bgd.shape[2] == 3:
        r_perp_x = ts_vec_xyz(b_bgd.time.data, b_bgd.data[:, 0])
        r_perp_y = ts_vec_xyz(b_bgd.time.data, b_bgd.data[:, 1])
        r_para = ts_vec_xyz(b_bgd.time.data, b_bgd.data[:, 2])
    else:
        raise TypeError("b_bgd must be a vector or a tensor time series")

    if inp_data.ndim == 2 and inp_data.shape[1] == 3:
        out_data = np.zeros(inp.shape, dtype=inp_data.dtype)

        out_data[:, 0] = np.sum(r_perp_x * inp_data, axis=1)
        out_data[:, 1] = np.sum(r_perp_y * inp_data, axis=1)
        out_data[:, 2] = np.sum(r_para * inp_data, axis=1)

        # To xarray
        out = ts_vec_xyz(time, out_data, attrs=inp.attrs)

    elif inp_data.ndim == 1:
        out_data = np.zeros([inp_data.shape[0], 2])

        out_data[:, 0] = inp * np.sum(r_perp_x * r_xyz, axis=1)
        out_data[:, 1] = inp * np.sum(r_para * r_xyz, axis=1)

        out = xr.DataArray(
            out_data, coords=[time, ["perp", "para"]], dims=["time", "comp"]
        )
    else:
        raise TypeError("inp must be a vector or scalar")

    return out
